"""
=======================
MoTIF on simulated data
=======================

This example demonstrates MoTIF [1] on simulated data. In the
alphacsc module, we are offering all the alternatives for the users
to try. Please cite our paper [2] if you use this implementation.

[1] Jost, Philippe, et al.
    "MoTIF: An efficient algorithm for learning
    translation invariant dictionaries." Acoustics, Speech and Signal
    Processing, 2006. ICASSP 2006 Proceedings. 2006 IEEE International
    Conference on. Vol. 5. IEEE, 2006.
[2] Jas, M., Dupr'e La Tour, T., Simsekli, U., & Gramfort, A. (2017).
    Learning the Morphology of Brain Signals Using Alpha-Stable Convolutional
    Sparse Coding. arXiv preprint arXiv:1705.08006.
"""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first import the modules.

import matplotlib.pyplot as plt
from alphacsc.simulate import simulate_data
from alphacsc.other.motif import learn_atoms

###############################################################################
# and define the relevant parameters. Note we choose a large n_times
# to avoid overlapping atoms which MoTIF cannot handle

n_times_atom = 64  # L
n_times = 5000  # T
n_atoms = 2  # K
n_trials = 10  # N

###############################################################################
# simulate the data.

random_state_simulate = 1
X, ds_true, z_true = simulate_data(n_trials, n_times, n_times_atom,
                                   n_atoms, random_state_simulate,
                                   constant_amplitude=True)

n_times_atom = n_times_atom * 7  # XXX: hack
n_iter = 10
max_shift = 11  # after this, the algorithm breaks

###############################################################################
# Note, how we use constant_amplitude=True since
# MoTIF cannot handle atoms of varying amplitudes. Check out our examples
# on :ref:`vanilla CSC <sphx_glr_auto_examples_plot_simulate_csc.py>` and
# :ref:`alphaCSC <sphx_glr_auto_examples_plot_simulate_alphacsc.py>` to learn
# how to deal with such cases.
# Finally, let us estimate the atoms.

ds_hat = learn_atoms(X, n_atoms, n_times_atom, n_iter=n_iter,
                     max_shift=max_shift, random_state=42)
plt.plot(ds_hat.T)

###############################################################################
# Compare this to the original atoms
plt.figure()
plt.plot(ds_true.T)
plt.show()



import os.path
import numpy as np


home = os.path.expanduser("~")
data_dir = os.path.join(home, "sdtw_data")
ucr_dir = os.path.join(data_dir, "UCR_TS_Archive_2015")


def _parse_ucr(filename):
    y = []
    X = []
    for line in open(filename):
        line = line.strip()
        arr = line.split(",")
        label = int(arr[0])
        feat = list(map(float, arr[1:]))
        feat = np.array(feat).reshape(-1, 1)
        y.append(label)
        X.append(feat)
    return X, np.array(y)


def list_ucr():
    return sorted(os.listdir(ucr_dir))


def load_ucr(name):
    folder = os.path.join(ucr_dir, name)
    tr = os.path.join(folder, "%s_TRAIN" % name)
    te = os.path.join(folder, "%s_TEST" % name)

    try:
        X_tr, y_tr = _parse_ucr(tr)
        X_te, y_te = _parse_ucr(te)
    except IOError:
        raise IOError("Please copy UCR_TS_Archive_2015/ to $HOME/sdtw_data/. "
                      "Download from www.cs.ucr.edu/~eamonn/time_series_data.")

    y_tr = np.array(y_tr)
    y_te = np.array(y_te)
    X_tr = np.array(X_tr)
    X_te = np.array(X_te)

    return X_tr, y_tr, X_te, y_te



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp
cnp.import_array()


cdef compute_ztz_lil(object[:] z_i_data,
                     object[:] z_i_rows,
                     int nnz,
                     int n_times_valid,
                     cnp.ndarray[double, ndim=3] ztz):

    cdef int k, k0, ctk, ltk, t, t0
    cdef double z0, z
    cdef int n_atoms = len(z_i_data)
    cdef int n_times_atom = (ztz.shape[2] + 1) // 2

    lower_tk = np.zeros(n_atoms, dtype=int)
    current_tk = np.zeros(n_atoms, dtype=int)
    for _ in range(nnz):
        # Find the argmin of the curent tk
        k0 = 0
        t0 = n_times_valid
        for k, ctk in enumerate(current_tk):
            if ctk >= len(z_i_rows[k]):
                continue
            t = z_i_rows[k][ctk]
            if t < t0:
                k0 = k
                t0 = t
                z0 = z_i_data[k][ctk]

        # Update the pointers on lower valid indices
        for k in range(n_atoms):
            tk = z_i_rows[k]
            while (lower_tk[k] < len(tk) and
                    tk[lower_tk[k]] <= t0 - n_times_atom):
                lower_tk[k] += 1

        # compute the correlation from z_k[t] with z
        for k, (zk, tk, ltk) in enumerate(zip(z_i_data, z_i_rows, lower_tk)):
            for z, t in zip(zk[ltk:], tk[ltk:]):
                if t - t0 > n_times_atom - 1:
                    break
                ztz[k, k0, n_times_atom - 1 + t0 - t] += z0 * z

        current_tk[k0] += 1

    return ztz



cdef compute_ztz_csr(cnp.ndarray[double, ndim=1] z_i_data,
                     cnp.ndarray[int, ndim=1] z_i_indices,
                     cnp.ndarray[int, ndim=1] z_i_indptr,
                     cnp.ndarray[double, ndim=3] ztz):

    cdef int k, k0, i, i0, t, t0, start_k0, start_k, end_k0, end_k, idx
    cdef int n_atoms = ztz.shape[0]
    cdef int n_times_atom = (ztz.shape[2] + 1) // 2

    for k0 in range(n_atoms):
        start_k0 = z_i_indptr[k0]
        end_k0 = z_i_indptr[k0 + 1]
        for k in range(n_atoms):
            start_k = z_i_indptr[k]
            end_k = z_i_indptr[k + 1]
            for i0 in range(start_k0, end_k0):
                t0 = z_i_indices[i0]
                for i in range(start_k, end_k):
                    t = z_i_indices[i]
                    idx = n_times_atom - 1 + t0 - t
                    if 0 <= idx < 2 * n_times_atom - 1:
                        ztz[k, k0, idx] += z_i_data[i] * z_i_data[i0]
    return ztz


def _fast_compute_ztz_lil(z, n_times_atom):
    """
    ztz.shape = n_atoms, n_atoms, 2 * n_times_atom - 1
    z = list of LIL matrices of shape = (n_atoms, n_times - n_times_atom + 1)
    """
    n_trials = len(z)
    n_atoms, n_times_valid = z[0].shape
    ztz = np.zeros(shape=(n_atoms, n_atoms, 2 * n_times_atom - 1))
    for z_i in z:
        ztz = compute_ztz_lil(z_i.data, z_i.rows, z_i.nnz, n_times_valid, ztz)

    return ztz


def _fast_compute_ztz_csr(z, n_times_atom):
    """
    ztz.shape = n_atoms, n_atoms, 2 * n_times_atom - 1
    z = list of LIL matrices of shape = (n_atoms, n_times - n_times_atom + 1)
    """
    n_trials = len(z)
    n_atoms, n_times_valid = z[0].shape
    ztz = np.zeros(shape=(n_atoms, n_atoms, 2 * n_times_atom - 1))
    for z_i in z:
        z_i_csr = z_i.tocsr()
        ztz = compute_ztz_csr(z_i_csr.data, z_i_csr.indices, z_i_csr.indptr, ztz)

    return ztz


_fast_compute_ztz = _fast_compute_ztz_csr




try:
    from .sparse_conv import _fast_sparse_convolve_multi
    from .sparse_conv import _fast_sparse_convolve_multi_uv
    from .compute_ztz import _fast_compute_ztz
    from .compute_ztz import _fast_compute_ztz_lil, _fast_compute_ztz_csr
    from .compute_ztX import _fast_compute_ztX
    from .coordinate_descent import subtract_zhat_to_beta, update_dz_opt
    _CYTHON_AVAILABLE = True

    __all__ = ["_fast_sparse_convolve_multi", "_fast_sparse_convolve_multi_uv",
               "_fast_compute_ztz", "_fast_compute_ztz_lil",
               "_fast_compute_ztz_csr", "_fast_compute_ztX",
               "subtract_zhat_to_beta", "update_dz_opt"]

except ImportError:  # pragma: no cover
    _CYTHON_AVAILABLE = False


def _assert_cython():
    if not _CYTHON_AVAILABLE:  # pragma: no cover
        raise NotImplementedError("cython is a required dependency for this "
                                  "part of the code.")



import numpy as np
from scipy.signal import hilbert


def fast_hilbert(array):
    n_points = array.shape[0]
    n_fft = next_power2(n_points)
    return hilbert(array, n_fft)[:n_points]


def next_power2(num):
    """Compute the smallest power of 2 >= to num.(float -> int)"""
    return 2 ** int(np.ceil(np.log2(num)))



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

import numpy as np
from scipy.stats import levy_stable

from .utils import check_random_state


def estimate_phi_mh(X, Xhat, alpha, Phi, n_iter_mcmc, n_burnin_mcmc,
                    random_state, return_loglk=False, verbose=10):
    """Estimate the expectation of 1/phi by Metropolis-Hastings"""

    if n_iter_mcmc <= n_burnin_mcmc:
        raise ValueError('n_iter_mcmc must be greater than n_burnin_mcmc')

    n_trials, n_times = X.shape
    residual = (X - Xhat) ** 2

    tau = np.zeros((n_trials, n_times))
    rng = check_random_state(random_state)

    if return_loglk:
        loglk_all = np.zeros((n_iter_mcmc, 1))

    for i in range(n_iter_mcmc):
        scale = 2 * np.cos(np.pi * alpha / 4) ** (2 / alpha)
        Phi_p = levy_stable.rvs(alpha / 2, 1, loc=0, scale=scale,
                                size=(n_trials, n_times), random_state=rng)
        log_acc = 0.5 * np.log(Phi / Phi_p) + residual * (1 / Phi - 1 / Phi_p)
        log_u = np.log(rng.uniform(size=(n_trials, n_times)))
        ix = (log_acc > log_u)

        Phi[ix] = Phi_p[ix]

        if return_loglk:
            loglk = np.sum(-0.5 * np.log(Phi) - 0.5 * residual / Phi)
            loglk_all[i] = loglk
            if verbose > 5:
                print("Iter: %d\t loglk:%E\t NumAcc:%d" %
                      (i, loglk, np.sum(ix)))

        if (i >= n_burnin_mcmc):
            tau += 1 / Phi

    tau = tau / (n_iter_mcmc - n_burnin_mcmc)

    if return_loglk:
        return Phi, tau, loglk_all
    return Phi, tau



# flake8: noqa F401
from .dictionary import get_D, get_uv
from .convolution import construct_X, construct_X_multi, _choose_convolve
from .validation import check_random_state, check_consistent_shape
from .profile_this import profile_this



"""
==============================
Selecting random state for CSC
==============================

The CSC problem is non-convex. Therefore, the solution depends
on the initialization. Here, we show how to select the
best atoms amongst different initializations.

"""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# As before, let us first define the parameters of our model.

n_times_atom = 64  # L
n_times = 512  # T
n_atoms = 2  # K
n_trials = 100  # N
n_iter = 50

reg = 0.1

###############################################################################
# Here, we simulate the data

from alphacsc.simulate import simulate_data # noqa
from scipy.stats import levy_stable # noqa
from alphacsc import check_random_state # noqa

random_state_simulate = 1
X, ds_true, z_true = simulate_data(n_trials, n_times, n_times_atom,
                                   n_atoms, random_state_simulate)

# Add stationary noise:
fraction_corrupted = 0.02
n_corrupted_trials = int(fraction_corrupted * n_trials)

rng = check_random_state(random_state_simulate)
X += 0.01 * rng.randn(*X.shape)

idx_corrupted = rng.randint(0, n_trials,
                            size=n_corrupted_trials)

###############################################################################
# Now, we run vanilla CSC on the data but with different initializations.

from alphacsc import learn_d_z # noqa

pobjs, d_hats = list(), list()
for random_state in range(5):
    print('\nRandom state: %d' % random_state)
    pobj, times, d_hat, z_hat, reg = learn_d_z(
        X, n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
        solver_d_kwargs=dict(factr=100), random_state=random_state,
        n_jobs=1, verbose=1)
    pobjs.append(pobj[-1])
    d_hats.append(d_hat)

###############################################################################
# As we loop through the random states, we save the objective value `pobj`
# at the last iteration of the algorithm.
#
# Now, let us look at the atoms for different initializations.
import matplotlib.pyplot as plt # noqa
fig, axes = plt.subplots(1, 5, figsize=(17, 3), sharex=True, sharey=True)
for ax, this_pobjs, d_hat in zip(axes, pobjs, d_hats):
    ax.plot(d_hat.T)
    ax.plot(ds_true.T, 'k--')
    ax.set_title('pobj: %0.2f' % this_pobjs)

###############################################################################
# Note that lower the objective value, the better is the recovered atom.
# This is one reason why using a concrete mathematical objective function as in
# convolutional sparse coding is superior to heuristic methods.
# Now, we select the best atom amongst them.
import numpy as np # noqa

plt.figure()
plt.plot(d_hats[np.argmin(pobjs)].T)
plt.plot(ds_true.T, 'k--')
plt.show()



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

import numpy as np

from .utils import check_random_state, construct_X


def simulate_data(n_trials, n_times, n_times_atom, n_atoms, random_state=42,
                  constant_amplitude=False):
    """Simulate the data.

    Parameters
    ----------
    n_trials : int
        Number of samples / trials.
    n_times : int
        Number of time points.
    n_times_atom : int
        Number of time points.
    n_atoms : int
        Number of atoms.
    random_state : int | None
        If integer, fix the random state.
    constant_amplitude : float
        If True, the activations have constant amplitude.

    Returns
    -------
    X : array, shape (n_trials, n_times)
        The data
    ds : array, shape (k, n_times_atom)
        The true atoms.
    z : array, shape (n_trials, n_times - n_times_atom + 1)
        The true codes.

    Note
    ----
    X will be non-zero from n_times_atom to n_times.
    """
    # add atoms
    rng = check_random_state(random_state)
    ds = np.zeros((n_atoms, n_times_atom))
    for idx, shape, n_cycles in cycler(n_atoms, n_times_atom):
        ds[idx, :] = get_atoms(shape, n_times_atom, n_cycles=n_cycles)
    ds /= np.linalg.norm(ds, axis=1)[:, None]

    z = get_activations(rng, (n_atoms, n_trials, n_times - n_times_atom + 1),
                        constant_amplitude=constant_amplitude)
    X = construct_X(z, ds)

    assert X.shape == (n_trials, n_times)
    assert z.shape == (n_atoms, n_trials, n_times - n_times_atom + 1)
    assert ds.shape == (n_atoms, n_times_atom)

    return X, ds, z


def cycler(n_atoms, n_times_atom):
    idx = 0
    for n_cycles in range(1, n_times_atom // 2):
        for shape in ['triangle', 'square', 'sin']:
            yield idx, shape, n_cycles
            idx += 1
            if idx >= n_atoms:
                break
        if idx >= n_atoms:
            break


def get_activations(rng, shape_z, constant_amplitude=False):
    starts = list()
    n_atoms, n_trials, n_times_valid = shape_z
    for idx in range(n_atoms):
        starts.append(rng.randint(low=0, high=n_times_valid,
                      size=(n_trials,)))
    # add activations
    z = np.zeros(shape_z)
    for i in range(n_trials):
        for k_idx, start in enumerate(starts):
            if constant_amplitude:
                randnum = 1.
            else:
                randnum = rng.uniform()
            z[k_idx, i, starts[k_idx][i]] = randnum
    return z


def get_atoms(shape, n_times_atom, zero_mean=True, n_cycles=1):
    if shape == 'triangle':
        ds = list()
        for idx in range(n_cycles):
            ds.append(np.linspace(0, 1, n_times_atom // (2 * n_cycles)))
            ds.append(ds[-1][::-1])
        d = np.hstack(ds)
        d = np.pad(d, (0, n_times_atom - d.shape[0]), 'constant')
    elif shape == 'square':
        ds = list()
        for idx in range(n_cycles):
            ds.append(0.5 * np.ones((n_times_atom // (2 * n_cycles))))
            ds.append(-ds[-1])
        d = np.hstack(ds)
        d = np.pad(d, (0, n_times_atom - d.shape[0]), 'constant')
    elif shape == 'sin':
        d = np.sin(2 * np.pi * n_cycles * np.linspace(0, 1, n_times_atom))
    elif shape == 'cos':
        d = np.cos(2 * np.pi * n_cycles * np.linspace(0, 1, n_times_atom))
    if zero_mean:
        d -= np.mean(d)

    return d



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
import time

import numpy as np
from scipy import linalg
from scipy import optimize, signal
from joblib import Parallel, delayed

from .utils.convolution import _choose_convolve
from .utils.optim import power_iteration
from .utils import check_consistent_shape


def update_z(X, ds, reg, z0=None, debug=False, parallel=None,
             solver='l-bfgs', b_hat_0=None, solver_kwargs=dict(),
             sample_weights=None):
    """Update Z using L-BFGS with positivity constraints

    Parameters
    ----------
    X : array, shape (n_trials, n_times)
        The data array
    ds : array, shape (n_atoms, n_times_atom)
        The atoms.
    reg : float
        The regularization constant
    z0 : None | array, shape (n_atoms, n_trials, n_times_valid)
        Init for z (can be used for warm restart).
    debug : bool
        If True, check the grad.
    parallel : instance of Parallel
        Context manager for running joblibs in a loop.
    solver : 'l-bfgs' | 'ista' | 'fista'
        The solver to use.
    b_hat_0 : array, shape ((n_times - n_times_atom + 1) * n_atoms)
        init vector for power_iteration with 'ista' solver
    solver_kwargs : dict
        Parameters for the solver
    sample_weights: array, shape (n_trials, n_times)
        Weights applied on the cost function.

    Returns
    -------
    z : array, shape (n_trials, n_times - n_times_atom + 1)
        The true codes.
    """
    n_trials, n_times = X.shape
    check_consistent_shape(X, sample_weights)
    n_atoms, n_times_atom = ds.shape
    n_times_valid = n_times - n_times_atom + 1

    # now estimate the codes
    my_update_z = delayed(_update_z_idx)
    if parallel is None:
        parallel = Parallel(n_jobs=1)
    else:
        assert parallel.n_jobs >= 1

    zhats = parallel(
        my_update_z(X, ds, reg, z0, i, debug, solver, b_hat_0, solver_kwargs,
                    sample_weights)
        for i in np.array_split(np.arange(n_trials), parallel.n_jobs))
    z_hat = np.vstack(zhats)

    z_hat2 = z_hat.reshape((n_trials, n_atoms, n_times_valid))
    z_hat2 = np.swapaxes(z_hat2, 0, 1)

    return z_hat2


def _fprime(ds, zi, Xi=None, sample_weights=None, reg=None, return_func=False):
    """np.dot(D.T, X[i] - np.dot(D, zi)) + reg

    Parameters
    ----------
    ds : array, shape (n_atoms, n_times_atom)
        The atoms
    zi : array, shape (n_atoms * n_times_valid)
        The activations
    Xi : array, shape (n_times, ) or None
        The data array for one trial
    sample_weights : array, shape (n_times, ) or None
        The sample weights for one trial
    reg : float or None
        The regularization constant
    return_func : boolean
        Returns also the objective function, used to speed up LBFGS solver

    Returns
    -------
    (func) : float
        The objective function
    grad : array, shape (n_atoms * n_times_valid)
        The gradient
    """
    n_atoms, n_times_atom = ds.shape
    zi_reshaped = zi.reshape((n_atoms, -1))
    Dzi = _choose_convolve(zi_reshaped, ds)
    if Xi is not None:
        Dzi -= Xi

    if sample_weights is not None:
        if return_func:
            # preserve Dzi, we don't want to apply the weights twice in func
            wDzi = sample_weights * Dzi
        else:
            Dzi *= sample_weights
            wDzi = Dzi
    else:
        wDzi = Dzi

    if return_func:
        func = 0.5 * np.dot(wDzi, Dzi.T)
        if reg is not None:
            func += reg * zi.sum()

    # Now do the dot product with the transpose of D (D.T) which is
    # the conv by the reversed filter (keeping valid mode)
    grad = np.concatenate(
        [signal.convolve(wDzi, d[::-1], 'valid') for d in ds])
    # grad = -np.dot(D.T, X[i] - np.dot(D, zi))
    if reg is not None:
        grad += reg

    if return_func:
        return func, grad
    else:
        return grad


def _update_z_idx(X, ds, reg, z0, idxs, debug, solver='l-bfgs', b_hat_0=None,
                  solver_kwargs=dict(), sample_weights=None, timing=False):

    n_trials, n_times = X.shape
    n_atoms, n_times_atom = ds.shape
    n_times_valid = n_times - n_times_atom + 1
    bounds = [(0, None) for idx in range(n_atoms * n_times_valid)]

    zhats = []

    for i in idxs:
        if sample_weights is None:
            sample_weights_i = None
        else:
            sample_weights_i = sample_weights[i]

        def func_and_grad(zi):
            return _fprime(ds, zi, Xi=X[i], reg=reg, return_func=True,
                           sample_weights=sample_weights_i)

        def grad_noreg(zi):
            return _fprime(ds, zi, Xi=X[i], reg=None, return_func=False,
                           sample_weights=sample_weights_i)

        if z0 is None:
            f0 = np.zeros(n_atoms * n_times_valid)
        else:
            f0 = z0[:, i, :].reshape((n_atoms * n_times_valid))

        if timing:
            times = [0]
            pobj = [func_and_grad(f0)[0]]
            start = [time.time()]

        if debug:

            def pobj(zi):
                return func_and_grad(zi)[0]

            def fprime(zi):
                return func_and_grad(zi)[1]

            assert optimize.check_grad(pobj, fprime, f0) < 1e-5

        if solver == 'l-bfgs':
            if timing:
                def callback(xk):
                    times.append(time.time() - start[0])
                    pobj.append(func_and_grad(xk)[0])
                    # use a reference to have access inside this function
                    start[0] = time.time()

            else:
                callback = None
            factr = solver_kwargs.get('factr', 1e15)  # default value
            maxiter = solver_kwargs.get('maxiter', 15000)  # default value
            zhat, f, d = optimize.fmin_l_bfgs_b(func_and_grad, f0, fprime=None,
                                                args=(), approx_grad=False,
                                                bounds=bounds, factr=factr,
                                                maxiter=maxiter,
                                                callback=callback)
        elif solver == "ista":
            zhat = f0.copy()
            DTD = gram_block_circulant(ds, n_times_valid, 'custom',
                                       sample_weights=sample_weights_i)
            tol = solver_kwargs.get('power_iteration_tol', 1e-4)
            L = power_iteration(DTD, b_hat_0=b_hat_0, tol=tol)
            step_size = 0.99 / L

            max_iter = solver_kwargs.get('max_iter', 20)
            for k in range(max_iter):  # run ISTA iterations
                zhat -= step_size * grad_noreg(zhat)
                zhat = np.maximum(zhat - reg * step_size, 0.)

                if timing:
                    times.append(time.time() - start[0])
                    pobj.append(func_and_grad(zhat)[0])
                    start[0] = time.time()

        elif solver == "fista":
            # init
            x_new = f0.copy()
            y = x_new.copy()
            t_new = 1.0

            DTD = gram_block_circulant(ds, n_times_valid, 'custom',
                                       sample_weights=sample_weights_i)
            # compute the Lipschitz constant
            tol = solver_kwargs.get('power_iteration_tol', 1e-4)
            L = power_iteration(DTD, b_hat_0=b_hat_0, tol=tol)
            step_size = 0.99 / L

            max_iter = solver_kwargs.get('max_iter', 20)
            restart = solver_kwargs.get('restart', None)
            for k in range(max_iter):  # run FISTA iterations
                # restart every n iterations
                if k > 0 and restart is not None and (k % restart) == 0:
                    y = x_new.copy()
                    t_new = 1.0

                # update the old
                t_old = t_new
                x_old = x_new

                # update x
                y -= step_size * grad_noreg(y)
                x_new = np.maximum(y - reg * step_size, 0.)

                # update t and y
                t_new = 0.5 * (1. + np.sqrt(1. + 4. * (t_old ** 2)))
                y = x_new + ((t_old - 1.) / t_new) * (x_new - x_old)

                if timing:
                    times.append(time.time() - start[0])
                    pobj.append(func_and_grad(x_new)[0])
                    start[0] = time.time()

            zhat = x_new
        else:
            raise ValueError("Unrecognized solver %s. Must be 'ista', 'fista',"
                             " or 'l-bfgs'." % solver)

        zhats.append(zhat)
    if timing:
        return np.vstack(zhats), pobj, times
    return np.vstack(zhats)


def gram_block_circulant(ds, n_times_valid, method='full',
                         sample_weights=None):
    """Returns ...

    Parameters
    ----------
    ds : array, shape (n_atoms, n_times_atom)
        The atoms
    n_times_valid : int
        n_times - n_times_atom + 1
    method : string
        If 'full', returns full circulant matrix.
        If 'scipy', returns scipy linear operator.
        If 'custom', returns custom linear operator.
    sample_weights : array, shape (n_times, )
        The sample weights for one trial
    """
    from scipy.sparse.linalg import LinearOperator
    from functools import partial

    n_atoms, n_times_atom = ds.shape
    n_times = n_times_valid + n_times_atom - 1

    if method == 'full':
        D = np.zeros((n_times, n_atoms * n_times_valid))
        for k_idx in range(n_atoms):
            d_padded = np.zeros((n_times, ))
            d_padded[:n_times_atom] = ds[k_idx]
            start = k_idx * n_times_valid
            stop = start + n_times_valid
            D[:, start:stop] = linalg.circulant((d_padded))[:, :n_times_valid]
        if sample_weights is not None:
            wD = sample_weights[:, None] * D
            return np.dot(D.T, wD)
        else:
            return np.dot(D.T, D)

    elif method == 'scipy':

        def matvec(v, ds):
            assert v.shape[0] % ds.shape[0] == 0
            return _fprime(ds, v, Xi=None, reg=None,
                           sample_weights=sample_weights)

        D = LinearOperator((n_atoms * n_times_valid, n_atoms * n_times_valid),
                           matvec=partial(matvec, ds=ds))

    elif method == 'custom':
        return CustomLinearOperator(ds, n_times_valid, sample_weights)
    else:
        raise ValueError('Unkown method %s.' % method)
    return D


class CustomLinearOperator():
    """Simpler class than scipy's LinearOperator, with less overhead

    Parameters
    ----------
    ds : array, shape (n_atoms, n_times_atom)
        The atoms
    n_times_valid : int
        n_times - n_times_atom + 1
    sample_weights : array, shape (n_times, )
        The sample weights for one trial
    """

    def __init__(self, ds, n_times_valid, sample_weights):
        self.ds = ds
        product = ds.shape[0] * n_times_valid
        self.shape = (product, product)
        self.sample_weights = sample_weights

    def dot(self, v):
        ds = self.ds
        assert v.shape[0] % ds.shape[0] == 0
        return _fprime(ds, v, Xi=None, reg=None,
                       sample_weights=self.sample_weights)



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>

import numpy as np

from .utils.convolution import numpy_convolve_uv
from .utils.convolution import tensordot_convolve
from .utils.convolution import _choose_convolve_multi
from .utils.whitening import apply_whitening
from .utils.lil import scale_z_by_atom, safe_sum, get_z_shape, is_list_of_lil
from .utils import construct_X_multi

try:
    from .other import sdtw as _sdtw
except ImportError:
    _sdtw = None


def _assert_dtw():
    if _sdtw is None:
        raise NotImplementedError("Could not import alphacsc.other.sdtw. This "
                                  "module must be compiled with cython to "
                                  "make loss='dtw' available.")


def compute_objective(X=None, X_hat=None, z_hat=None, D=None,
                      constants=None, reg=None, loss='l2', loss_params=dict()):
    """Compute the value of the objective function

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    X_hat : array, shape (n_trials, n_channels, n_times)
        The current reconstructed signal.
    z_hat : array, shape (n_atoms, n_trials, n_times_valid)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The current activation signals for the regularization.
    constants : dict
        Constant to accelerate the computation when updating uv.
    reg : float or array, shape (n_atoms, )
        The regularization parameters. If None, no regularization is added.
        The regularization constant
    loss : str in {'l2' | 'dtw'}
        Loss function for the data-fit
    loss_params : dict
        Parameter for the loss
    """
    if loss == 'l2':
        obj = _l2_objective(X=X, X_hat=X_hat, D=D, constants=constants)
    elif loss == 'dtw':
        _assert_dtw()
        obj = _dtw_objective(X, X_hat, loss_params=loss_params)
    elif loss == 'whitening':
        ar_model = loss_params['ar_model']

        # X is assumed to be already whitened, just select the valid part
        X = X[:, :, ar_model.ordar:-ar_model.ordar]
        X_hat = apply_whitening(ar_model, X_hat, mode='valid')
        obj = _l2_objective(X=X, X_hat=X_hat, D=D, constants=constants)
    else:
        raise NotImplementedError("loss '{}' is not implemented".format(loss))

    if reg is not None:
        if isinstance(reg, (int, float)):
            obj += reg * safe_sum(z_hat)
        else:
            obj += np.sum(reg * safe_sum(z_hat, axis=(1, 2)))

    return obj


def compute_X_and_objective_multi(X, z_hat, D_hat=None, reg=None, loss='l2',
                                  loss_params=dict(), feasible_evaluation=True,
                                  uv_constraint='joint', return_X_hat=False):
    """Compute X and return the value of the objective function

    Parameters
    ----------
    X : array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    z_hat : array, shape (n_trials, n_atoms, n_times - n_times_atom + 1)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The sparse activation matrix.
    uv_hat : array, shape (n_atoms, n_channels + n_times_atom)
        The atoms to learn from the data.
    reg : float
        The regularization Parameters
    loss : 'l2' | 'dtw'
        Loss to measure the discrepency between the signal and our estimate.
    loss_params : dict
        Parameters of the loss
    feasible_evaluation: boolean
        If feasible_evaluation is True, it first projects on the feasible set,
        i.e. norm(uv_hat) <= 1.
    uv_constraint : str in {'joint', 'separate'}
        The kind of norm constraint on the atoms:
        If 'joint', the constraint is norm([u, v]) <= 1
        If 'separate', the constraint is norm(u) <= 1 and norm(v) <= 1
    """
    n_channels = X.shape[1]

    if feasible_evaluation:
        if D_hat.ndim == 2:
            D_hat = D_hat.copy()
            # project to unit norm
            from .update_d_multi import prox_uv
            D_hat, norm = prox_uv(D_hat, uv_constraint=uv_constraint,
                                  n_channels=n_channels, return_norm=True)
        else:
            D_hat = D_hat.copy()
            # project to unit norm
            from .update_d_multi import prox_d
            D_hat, norm = prox_d(D_hat, return_norm=True)

        # update z in the opposite way
        z_hat = scale_z_by_atom(z_hat, scale=norm, copy=True)

    X_hat = construct_X_multi(z_hat, D=D_hat, n_channels=n_channels)

    cost = compute_objective(X=X, X_hat=X_hat, z_hat=z_hat, reg=reg, loss=loss,
                             loss_params=loss_params)
    if return_X_hat:
        return cost, X_hat
    return cost


def compute_gradient_norm(X, z_hat, D_hat, reg, loss='l2', loss_params=dict(),
                          rank1=False, sample_weights=None):
    if X.ndim == 2:
        X = X[:, None, :]
        D_hat = D_hat[:, None, :]

    if rank1:
        grad_d = gradient_uv(uv=D_hat, X=X, z=z_hat, constants=None,
                             loss=loss, loss_params=loss_params)
    else:
        grad_d = gradient_d(D=D_hat, X=X, z=z_hat, constants=None,
                            loss=loss, loss_params=loss_params)

    grad_norm_z = 0
    for i in range(X.shape[0]):
        grad_zi = gradient_zi(X[i], z_hat[i], D=D_hat, reg=reg,
                              loss=loss, loss_params=loss_params)
        grad_norm_z += np.dot(grad_zi.ravel(), grad_zi.ravel())

    grad_norm_d = np.dot(grad_d.ravel(), grad_d.ravel())
    grad_norm = np.sqrt(grad_norm_d) + np.sqrt(grad_norm_z)

    return grad_norm


def gradient_uv(uv, X=None, z=None, constants=None, reg=None, loss='l2',
                loss_params=dict(), return_func=False, flatten=False):
    """Compute the gradient of the reconstruction loss relative to uv.

    Parameters
    ----------
    uv : array, shape (n_atoms, n_channels + n_times_atom)
        The spatial and temporal atoms
    X : array, shape (n_trials, n_channels, n_times) or None
        The data array
    z : array, shape (n_atoms, n_trials, n_times_valid) or None
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The activations
    constants : dict or None
        Constant to accelerate the computation of the gradient
    reg : float or None
        The regularization constant
    loss : str in {'l2' | 'dtw'}
        Loss function for the data-fit
    loss_params : dict
        Parameter for the loss
    return_func : boolean
        Returns also the objective function, used to speed up LBFGS solver
    flatten : boolean
        If flatten is True, takes a flatten uv input and return the gradient
        as a flatten array.

    Returns
    -------
    (func) : float
        The objective function
    grad : array, shape (n_atoms * n_times_valid)
        The gradient
    """
    if z is not None:
        assert X is not None
        n_atoms = get_z_shape(z)[1]
        n_channels = X.shape[1]
    else:
        n_atoms = constants['ztz'].shape[0]
        n_channels = constants['n_channels']

    if is_list_of_lil(z) and loss != 'l2':
        raise NotImplementedError()

    if flatten:
        uv = uv.reshape((n_atoms, -1))

    if loss == 'l2':
        cost, grad_d = _l2_gradient_d(D=uv, X=X, z=z, constants=constants)
    elif loss == 'dtw':
        _assert_dtw()
        cost, grad_d = _dtw_gradient_d(D=uv, X=X, z=z, loss_params=loss_params)
    elif loss == 'whitening':
        cost, grad_d = _whitening_gradient_d(D=uv, X=X, z=z,
                                             loss_params=loss_params)
    else:
        raise NotImplementedError("loss {} is not implemented.".format(loss))
    grad_u = (grad_d * uv[:, None, n_channels:]).sum(axis=2)
    grad_v = (grad_d * uv[:, :n_channels, None]).sum(axis=1)
    grad = np.c_[grad_u, grad_v]

    if flatten:
        grad = grad.ravel()

    if return_func:
        if reg is not None:
            if isinstance(reg, float):
                cost += reg * safe_sum(z)
            else:
                cost += np.sum(reg * safe_sum(z, axis=(1, 2)))
        return cost, grad

    return grad


def gradient_zi(Xi, zi, D=None, constants=None, reg=None, loss='l2',
                loss_params=dict(), return_func=False, flatten=False):
    n_atoms = D.shape[0]
    if flatten:
        zi = zi.reshape((n_atoms, -1))

    if loss == 'l2':
        cost, grad = _l2_gradient_zi(Xi, zi, D=D, return_func=return_func)
    elif loss == 'dtw':
        _assert_dtw()
        cost, grad = _dtw_gradient_zi(Xi, zi, D=D, loss_params=loss_params)
    elif loss == 'whitening':
        cost, grad = _whitening_gradient_zi(Xi, zi, D=D,
                                            loss_params=loss_params,
                                            return_func=return_func)
    else:
        raise NotImplementedError("loss {} is not implemented.".format(loss))

    if reg is not None:
        grad += reg
        if return_func:
            if isinstance(reg, float):
                cost += reg * zi.sum()
            else:
                cost += np.sum(reg * zi.sum(axis=1))

    if flatten:
        grad = grad.ravel()

    if return_func:
        return cost, grad

    return grad


def gradient_d(D=None, X=None, z=None, constants=None, reg=None,
               loss='l2', loss_params=dict(), return_func=False,
               flatten=False):
    """Compute the gradient of the reconstruction loss relative to d.

    Parameters
    ----------
    D : array
        The atoms. Can either be full rank with shape shape
        (n_atoms, n_channels, n_times_atom) or rank 1 with
        shape shape (n_atoms, n_channels + n_times_atom)
    X : array, shape (n_trials, n_channels, n_times) or None
        The data array
    z : array, shape (n_atoms, n_trials, n_times_valid) or None
        The activations
    constants : dict or None
        Constant to accelerate the computation of the gradient
    reg : float or None
        The regularization constant
    loss : str in {'l2' | 'dtw'}
        Loss function for the data-fit
    loss_params : dict
        Parameter for the loss
    return_func : boolean
        Returns also the objective function, used to speed up LBFGS solver
    flatten : boolean
        If flatten is True, takes a flatten uv input and return the gradient
        as a flatten array.

    Returns
    -------
    (func) : float
        The objective function
    grad : array, shape (n_atoms * n_times_valid)
        The gradient
    """
    if flatten:
        if z is None:
            n_atoms = constants['ztz'].shape[0]
            n_channels = constants['n_channels']
        else:
            n_atoms = get_z_shape(z)[1]
            n_channels = X.shape[1]
        D = D.reshape((n_atoms, n_channels, -1))

    if is_list_of_lil(z) and loss != 'l2':
        raise NotImplementedError()

    if loss == 'l2':
        cost, grad_d = _l2_gradient_d(D=D, X=X, z=z, constants=constants)
    elif loss == 'dtw':
        _assert_dtw()
        cost, grad_d = _dtw_gradient_d(D=D, X=X, z=z, loss_params=loss_params)
    elif loss == 'whitening':
        cost, grad_d = _whitening_gradient_d(D=D, X=X, z=z,
                                             loss_params=loss_params)
    else:
        raise NotImplementedError("loss {} is not implemented.".format(loss))

    if flatten:
        grad_d = grad_d.ravel()

    if return_func:
        if reg is not None:
            if isinstance(reg, float):
                cost += reg * safe_sum(z)
            else:
                cost += np.dot(reg, safe_sum(z, axis=(1, 2)))
        return cost, grad_d

    return grad_d


def _dtw_objective(X, X_hat, loss_params=dict()):
    gamma = loss_params.get('gamma')
    sakoe_chiba_band = loss_params.get('sakoe_chiba_band', -1)

    n_trials = X.shape[0]
    cost = 0
    for idx in range(n_trials):
        D_X = _sdtw.distance.SquaredEuclidean(X_hat[idx].T, X[idx].T)
        sdtw = _sdtw.SoftDTW(D_X, gamma=gamma,
                             sakoe_chiba_band=sakoe_chiba_band)
        cost += sdtw.compute()

    return cost


def _dtw_gradient(X, z, D=None, loss_params=dict()):
    gamma = loss_params.get('gamma')
    sakoe_chiba_band = loss_params.get('sakoe_chiba_band', -1)

    n_trials, n_channels, n_times = X.shape
    X_hat = construct_X_multi(z, D=D, n_channels=n_channels)
    grad = np.zeros(X_hat.shape)
    cost = 0
    for idx in range(n_trials):
        D_X = _sdtw.distance.SquaredEuclidean(X_hat[idx].T, X[idx].T)
        sdtw = _sdtw.SoftDTW(D_X, gamma=gamma,
                             sakoe_chiba_band=sakoe_chiba_band)

        cost += sdtw.compute()
        grad[idx] = D_X.jacobian_product(sdtw.grad()).T

    return cost, grad


def _dtw_gradient_d(D, X=None, z=None, loss_params={}):
    cost, grad_X_hat = _dtw_gradient(X, z, D=D, loss_params=loss_params)

    return cost, _dense_transpose_convolve_z(grad_X_hat, z)


def _dtw_gradient_zi(Xi, z_i, D=None, loss_params={}):
    n_channels = Xi.shape[0]
    cost_i, grad_Xi_hat = _dtw_gradient(Xi[None], z_i[None], D=D,
                                        loss_params=loss_params)

    return cost_i, _dense_transpose_convolve_d(
        grad_Xi_hat[0], D=D, n_channels=n_channels)


def _l2_gradient_d(D, X=None, z=None, constants=None):

    if constants:
        assert D is not None
        if D.ndim == 2:
            g = numpy_convolve_uv(constants['ztz'], D)
        else:
            g = tensordot_convolve(constants['ztz'], D)
        return None, g - constants['ztX']
    else:
        n_channels = X.shape[1]
        residual = construct_X_multi(z, D=D, n_channels=n_channels) - X
        return None, _dense_transpose_convolve_z(residual, z)


def _l2_objective(X=None, X_hat=None, D=None, constants=None):

    if constants:
        # Fast compute the l2 objective when updating uv/D
        assert D is not None, "D is needed to fast compute the objective."
        if D.ndim == 2:
            # rank 1 dictionary, use uv computation
            n_channels = constants['n_channels']
            grad_d = .5 * numpy_convolve_uv(constants['ztz'], D)
            grad_d -= constants['ztX']
            cost = (grad_d * D[:, None, n_channels:]).sum(axis=2)
            cost = np.dot(cost.ravel(), D[:, :n_channels].ravel())
        else:
            grad_d = .5 * tensordot_convolve(constants['ztz'], D)
            grad_d -= constants['ztX']
            cost = (D * grad_d).sum()

        cost += .5 * constants['XtX']
        return cost

    # else, compute the l2 norm of the residual
    assert X is not None and X_hat is not None
    residual = X - X_hat
    return 0.5 * np.dot(residual.ravel(), residual.ravel())


def _l2_gradient_zi(Xi, z_i, D=None, return_func=False):
    """

    Parameters
    ----------
    Xi : array, shape (n_channels, n_times)
        The data array for one trial
    z_i : array, shape (n_atoms, n_times_valid)
        The activations
    D : array
        The current dictionary, it can have shapes:
        - (n_atoms, n_channels + n_times_atom) for rank 1 dictionary
        - (n_atoms, n_channels, n_times_atom) for full rank dictionary
    return_func : boolean
        Returns also the objective function, used to speed up LBFGS solver

    Returns
    -------
    (func) : float
        The objective function l2
    grad : array, shape (n_atoms, n_times_valid)
        The gradient
    """
    n_channels, _ = Xi.shape

    Dz_i = _choose_convolve_multi(z_i, D=D, n_channels=n_channels)

    # n_channels, n_times = Dz_i.shape
    if Xi is not None:
        Dz_i -= Xi

    if return_func:
        func = 0.5 * np.dot(Dz_i.ravel(), Dz_i.ravel())
    else:
        func = None

    grad = _dense_transpose_convolve_d(Dz_i, D=D, n_channels=n_channels)

    return func, grad


def _whitening_gradient(X, X_hat, loss_params, return_func=False):
    ar_model = loss_params['ar_model']

    # Xi is assumed to be already whitened, select the valid part
    X = X[:, :, ar_model.ordar:-ar_model.ordar]

    # Construct X_hat and whitten it
    X_hat = apply_whitening(ar_model, X_hat, mode='valid')
    residual = X_hat - X

    if return_func:
        func = 0.5 * np.dot(residual.ravel(), residual.ravel())
    else:
        func = None

    hTh_res = apply_whitening(ar_model, residual, reverse_ar=True,
                              mode='full')

    return hTh_res, func


def _whitening_gradient_zi(Xi, z_i, D, loss_params, return_func=False):
    n_channels, n_times = Xi.shape

    # Construct Xi_hat and compute the gradient relatively to X_hat
    Xi_hat = construct_X_multi(z_i[None], D=D, n_channels=n_channels)
    hTh_res, func = _whitening_gradient(Xi[None], Xi_hat, loss_params,
                                        return_func=return_func)

    # Use the chain rule to compute the gradient compared to z_i
    grad = _dense_transpose_convolve_d(hTh_res[0], D=D, n_channels=n_channels)
    assert grad.shape == z_i.shape

    return func, grad


def _whitening_gradient_d(D, X, z, loss_params):
    n_channels = X.shape[1]

    # Construct Xi_hat and compute the gradient relatively to X_hat
    X_hat = construct_X_multi(z, D=D, n_channels=n_channels)
    hTh_res, func = _whitening_gradient(X, X_hat, loss_params,
                                        return_func=False)

    # Use the chain rule to compute the gradient compared to D
    grad = _dense_transpose_convolve_z(hTh_res, z)

    return None, grad


def _dense_transpose_convolve_z(residual, z):
    """Convolve residual[i] with the transpose for each atom k, and return the sum

    Parameters
    ----------
    residual : array, shape (n_trials, n_channels, n_times)
    z : array, shape (n_trials, n_atoms, n_times_valid)

    Return
    ------
    grad_D : array, shape (n_atoms, n_channels, n_times_atom)

    """
    if is_list_of_lil(z):
        raise NotImplementedError()

    return np.sum([[[np.convolve(res_ip, z_ik[::-1],
                                 mode='valid')                   # n_times_atom
                     for res_ip in res_i]                        # n_channnels
                    for z_ik in z_i]                             # n_atoms
                   for z_i, res_i in zip(z, residual)], axis=0)  # n_atoms


def _dense_transpose_convolve_d(residual_i, D=None, n_channels=None):
    """Convolve residual[i] with the transpose for each atom k

    Parameters
    ----------
    residual_i : array, shape (n_channels, n_times)
    D : array, shape (n_atoms, n_channels, n_times_atom) or
               shape (n_atoms, n_channels + n_times_atom)

    Return
    ------
    grad_zi : array, shape (n_atoms, n_times_valid)

    """

    if D.ndim == 2:
        # multiply by the spatial filter u
        uR_i = np.dot(D[:, :n_channels], residual_i)

        # Now do the dot product with the transpose of D (D.T) which is
        # the conv by the reversed filter (keeping valid mode)
        return np.array([
            np.convolve(uR_ik, v_k[::-1], 'valid')
            for (uR_ik, v_k) in zip(uR_i, D[:, n_channels:])
        ])
    else:
        return np.sum([[np.correlate(res_ip, d_kp, mode='valid')
                        for res_ip, d_kp in zip(residual_i, d_k)]
                       for d_k in D], axis=1)



# Author: Mathieu Blondel
# License: Simplified BSD

import numpy as np

from scipy.optimize import minimize

from sdtw import SoftDTW
from sdtw.distance import SquaredEuclidean


def sdtw_barycenter(X, barycenter_init, gamma=1.0, weights=None,
                    method="L-BFGS-B", tol=1e-3, max_iter=50):
    """
    Compute barycenter (time series averaging) under the soft-DTW geometry.

    Parameters
    ----------
    X: list
        List of time series, numpy arrays of shape [len(X[i]), d].

    barycenter_init: array, shape = [length, d]
        Initialization.

    gamma: float
        Regularization parameter.
        Lower is less smoothed (closer to true DTW).

    weights: None or array
        Weights of each X[i]. Must be the same size as len(X).

    method: string
        Optimization method, passed to `scipy.optimize.minimize`.
        Default: L-BFGS.

    tol: float
        Tolerance of the method used.

    max_iter: int
        Maximum number of iterations.
    """
    if weights is None:
        weights = np.ones(len(X))

    weights = np.array(weights)

    def _func(Z):
        # Compute objective value and grad at Z.

        Z = Z.reshape(*barycenter_init.shape)

        G = np.zeros_like(Z)

        obj = 0

        for i in range(len(X)):
            D = SquaredEuclidean(Z, X[i])
            sdtw = SoftDTW(D, gamma=gamma)
            value = sdtw.compute()
            E = sdtw.grad()
            G_tmp = D.jacobian_product(E)
            G += weights[i] * G_tmp
            obj += weights[i] * value

        return obj, G.ravel()

    # The function works with vectors so we need to vectorize barycenter_init.
    res = minimize(_func, barycenter_init.ravel(), method=method, jac=True,
                   tol=tol, options=dict(maxiter=max_iter, disp=False))

    return res.x.reshape(*barycenter_init.shape)



"""
=====================================================================
Extracting artifact and evoked response atoms from the sample dataset
=====================================================================

This example illustrates how to learn rank-1 [1]_ atoms on the multivariate
sample dataset from :code:`mne`. We display a selection of atoms, featuring
heartbeat and eyeblink artifacts, three atoms of evoked responses, and a
non-sinusoidal oscillation.

.. [1] Dupré La Tour, T., Moreau, T., Jas, M., & Gramfort, A. (2018).
    `Multivariate Convolutional Sparse Coding for Electromagnetic Brain Signals
    <https://arxiv.org/abs/1805.09654v2>`_. Advances in Neural Information
    Processing Systems (NIPS).
"""

# Authors: Thomas Moreau <thomas.moreau@inria.fr>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first define the parameters of our model.

# sample frequency
sfreq = 150.

# Define the shape of the dictionary
n_atoms = 40
n_times_atom = int(round(sfreq * 1.0))  # 1000. ms

# Regularization parameter which control sparsity
reg = 0.1

# number of processors for parallel computing
n_jobs = 5

###############################################################################
# Next, we define the parameters for multivariate CSC

from alphacsc import GreedyCDL
cdl = GreedyCDL(
    # Shape of the dictionary
    n_atoms=n_atoms,
    n_times_atom=n_times_atom,
    # Request a rank1 dictionary with unit norm temporal and spatial maps
    rank1=True,
    uv_constraint='separate',
    # apply a temporal window reparametrization
    window=True,
    # at the end, refit the activations with fixed support and no reg to unbias
    unbiased_z_hat=True,
    # Initialize the dictionary with random chunk from the data
    D_init='chunk',
    # rescale the regularization parameter to be a percentage of lambda_max
    lmbd_max="scaled",
    reg=reg,
    # Number of iteration for the alternate minimization and cvg threshold
    n_iter=100,
    eps=1e-4,
    # solver for the z-step
    solver_z="lgcd",
    solver_z_kwargs={'tol': 1e-2,
                     'max_iter': 10000},
    # solver for the d-step
    solver_d='alternate_adaptive',
    solver_d_kwargs={'max_iter': 300},
    # sort atoms by explained variances
    sort_atoms=True,
    # Technical parameters
    verbose=1,
    random_state=0,
    n_jobs=n_jobs)


###############################################################################
# Here, we load the data from the sample datase. The data is not epoched yet,
# but we split it into 12 parts to make the most of multiple processors during
# the model fitting.

from alphacsc.datasets.somato import load_data
X_split, info = load_data(sfreq=sfreq, dataset='sample', n_splits=2 * n_jobs)

###############################################################################
# Fit the model and learn rank1 atoms
cdl.fit(X_split)

###############################################################################
# To avoid artifacts due to the splitting, we can optionally reload the data.
X, info = load_data(sfreq=sfreq, dataset='sample', n_splits=1)

###############################################################################
# Then we call the `transform` method, which returns the sparse codes
# associated with X, without changing the dictionary learned during the `fit`.
z_hat = cdl.transform(X)


###############################################################################
# Display a selection of atoms. We recognize a heartbeat artifact, an
# eyeblink artifact, three atoms of evoked responses, and a non-sinusoidal
# oscillation.

import mne
import numpy as np
import matplotlib.pyplot as plt

from alphacsc.utils.convolution import construct_X_multi
from alphacsc.viz.epoch import plot_evoked_surrogates

# preselected atoms of interest
plotted_atoms = [2, 0, 3, 15, 20, 11]

n_plots = 3  # number of plots by atom
n_columns = min(6, len(plotted_atoms))
split = int(np.ceil(len(plotted_atoms) / n_columns))
figsize = (4 * n_columns, 3 * n_plots * split)
fig, axes = plt.subplots(n_plots * split, n_columns, figsize=figsize)
for ii, kk in enumerate(plotted_atoms):

    # Select the axes to display the current atom
    print("\rDisplaying {}-th atom".format(kk), end='', flush=True)
    i_row, i_col = ii // n_columns, ii % n_columns
    it_axes = iter(axes[i_row * n_plots:(i_row + 1) * n_plots, i_col])

    # Select the current atom
    u_k = cdl.u_hat_[kk]
    v_k = cdl.v_hat_[kk]

    # Plot the spatial map of the atom using mne topomap
    ax = next(it_axes)
    mne.viz.plot_topomap(u_k, info, axes=ax, show=False)
    ax.set(title="Spatial pattern %d" % (kk, ))

    # Plot the temporal pattern of the atom
    ax = next(it_axes)
    t = np.arange(n_times_atom) / sfreq
    ax.plot(t, v_k)
    ax.set_xlim(0, n_times_atom / sfreq)
    ax.set(xlabel='Time (sec)', title="Temporal pattern %d" % kk)

    # Plot the power spectral density (PSD)
    ax = next(it_axes)
    psd = np.abs(np.fft.rfft(v_k, n=256)) ** 2
    frequencies = np.linspace(0, sfreq / 2.0, len(psd))
    ax.semilogy(frequencies, psd, label='PSD', color='k')
    ax.set(xlabel='Frequencies (Hz)', title="Power spectral density %d" % kk)
    ax.grid(True)
    ax.set_xlim(0, 30)
    ax.set_ylim(1e-4, 1e2)
    ax.legend()

fig.tight_layout()

###############################################################################
# Display the evoked reconstructed enveloppe:
# For each atom (columns), and for each event (rows), we compute the enveloppe
# of the reconstructed signal, align it with respect to the event onsets, and
# take the average. For some atoms, the activations are correlated with the
# events, leading to a large evoked enveloppe. The gray area corresponds to
# not statistically significant values.

from copy import deepcopy
from alphacsc.utils.signal import fast_hilbert

# time window around the events
t_lim = (-0.1, 0.5)

n_plots = len(np.atleast_1d(info['event_id']))
n_columns = min(6, len(plotted_atoms))
split = int(np.ceil(len(plotted_atoms) / n_columns))
figsize = (4 * n_columns, 3 * n_plots * split)
fig, axes = plt.subplots(n_plots * split, n_columns, figsize=figsize)

for ii, kk in enumerate(plotted_atoms):

    # Select the axes to display the current atom
    print("\rDisplaying {}-th atom".format(kk), end='', flush=True)
    i_row, i_col = ii // n_columns, ii % n_columns
    it_axes = iter(axes[i_row * n_plots:(i_row + 1) * n_plots, i_col])

    # Select the current atom
    v_k = cdl.v_hat_[kk]
    v_k_1 = np.r_[[1], v_k][None]
    z_k = z_hat[:, kk:kk + 1]
    X_k = construct_X_multi(z_k, v_k_1, n_channels=1)[0, 0]

    # compute the 'enveloppe' of the reconstructed signal X_k
    correlation = np.abs(fast_hilbert(X_k))

    # loop over all events IDs
    for this_event_id in np.atleast_1d(info['event_id']):
        # select the event by its ID, and create a new info dictionary
        this_info = deepcopy(info)
        this_info['events'] = this_info['events'][this_info['events'][:, 2] ==
                                                  this_event_id]
        this_info['event_id'] = this_event_id

        # plotting function
        ax = next(it_axes)
        plot_evoked_surrogates(correlation, info=this_info, t_lim=t_lim, ax=ax,
                               n_jobs=n_jobs, label='event %d' % this_event_id)
        ax.set(xlabel='Time (sec)', title="Evoked enveloppe %d" % kk)

fig.tight_layout()



"""
- This code implements a solver for the paper
  "Fast and Flexible Convolutional Sparse Coding" on signal data.
- The goal of this solver is to find the common filters, the codes for each
  signal series in the dataset
and a reconstruction for the dataset.
- The common filters (or kernels) and the codes for each image is denoted as
  d and z respectively
- We denote the step to solve the filter as d-step and the codes z-step
"""
import time
import numpy as np
from scipy import linalg, signal
from scipy.fftpack import fft, ifft
from mne.utils import check_random_state

real_type = 'float64'
imaginary_type = 'complex128'


def update_d(z_hat, d_hat, size_z, size_x, rho, d, v_D,
             d_D, lambdas, gammas_D, Mtb, u_D, M, size_k_full,
             psf_radius, xi_D, xi_D_hat, verbose, max_it_d):
    """D-STEP"""
    # Precompute what is necessary for later
    [zhat_mat, zhat_inv_mat] = precompute_D_step(z_hat, size_z, rho, verbose)

    for i_d in range(max_it_d):

        v_D[0] = np.real(
            ifft(np.einsum('ijk,jk->ik', z_hat, d_hat)))
        v_D[1] = d

        # Compute proximal updates
        u = v_D[0] - d_D[0]
        theta = lambdas[0] / gammas_D[0]
        u_D[0] = (Mtb + 1.0 / theta * u) / (M + 1.0 / theta * np.ones(size_x))

        u = v_D[1] - d_D[1]
        u_D[1] = KernelConstraintProj(u, size_k_full, psf_radius)

        # Update Langrange multipliers
        d_D[0] += u_D[0] - v_D[0]
        d_D[1] += u_D[1] - v_D[1]

        # Compute new xi=u+d and transform to fft
        xi_D[0] = u_D[0] + d_D[0]
        xi_D[1] = u_D[1] + d_D[1]
        xi_D_hat[0] = fft(xi_D[0])
        xi_D_hat[1] = fft(xi_D[1])

        # Solve convolutional inverse
        d_hat = solve_conv_term_D(
            zhat_mat, zhat_inv_mat, xi_D_hat, rho, size_z)
        d = np.real(ifft(d_hat))

    return d, d_hat


def update_z(z, z_hat, d_hat, u_Z, v_Z, d_Z, lambdas,
             gammas_Z, Mtb, M, size_x, size_z, xi_Z, xi_z_hat, b,
             lambda_prior, lambda_residual, psf_radius, verbose, max_it_z):
    """Z-STEP"""
    # Precompute what is necessary for later
    dhat_flat, dhatTdhat_flat = precompute_Z_step(d_hat, size_x, verbose)
    dhatT_flat = np.ma.conjugate(dhat_flat.T)

    for i_z in range(max_it_z):

        # Compute v = [Dz,z]
        v_Z[0] = np.real(
            ifft(np.einsum('ijk,jk->ik', z_hat, d_hat)))
        v_Z[1] = z

        # Compute proximal updates
        u = v_Z[0] - d_Z[0]
        theta = lambdas[0] / gammas_Z[0]
        u_Z[0] = (Mtb + 1.0 / theta * u) / (M + 1.0 / theta * np.ones(size_x))

        u = v_Z[1] - d_Z[1]
        theta = lambdas[1] / gammas_Z[1] * np.ones(u.shape)
        # u_Z[1] = np.multiply(np.maximum(
        #     0, 1 - np.divide(theta, np.abs(u))), u)
        u_Z[1] = np.maximum(u - theta, 0)  # positivity constraint

        # Update Lagrange multipliers
        d_Z[0] += u_Z[0] - v_Z[0]
        d_Z[1] += u_Z[1] - v_Z[1]

        # Compute new xi=u+d and transform to fft
        xi_Z[0] = u_Z[0] + d_Z[0]
        xi_Z[1] = u_Z[1] + d_Z[1]

        xi_z_hat[0] = fft(xi_Z[0])
        xi_z_hat[1] = fft(xi_Z[1])

        # Solve convolutional inverse
        z_hat = solve_conv_term_Z(
            dhatT_flat, dhatTdhat_flat, xi_z_hat, gammas_Z, size_z)
        z = np.real(ifft(z_hat))

    return z, z_hat


def learn_conv_sparse_coder(b, size_kernel, max_it, tol,
                            lambda_prior=1.0, lambda_residual=1.0,
                            random_state=None, ds_init=None,
                            feasible_evaluation=True,
                            stopping_pobj=None, verbose=1,
                            max_it_d=10, max_it_z=10):
    """
    Main function to solve the convolutional sparse coding.

    Parameters
    ----------
    - b               : the signal dataset with size (num_signals, length)
    - size_kernel     : the size of each kernel (num_kernels, length)
    - max_it          : the maximum iterations of the outer loop
    - tol             : the minimal difference in filters and codes after each
                        iteration to continue

    Important variables used in the code:
    - u_D, u_Z        : pair of proximal values for d-step and z-step
    - d_D, d_Z        : pair of Lagrange multipliers in the ADMM algo for
                        d-step and z-step
    - v_D, v_Z        : pair of initial value pairs (Zd, d) for d-step and
                        (Dz, z) for z-step
    """
    rng = check_random_state(random_state)

    k = size_kernel[0]
    n = b.shape[0]

    psf_radius = int(np.floor(size_kernel[1] / 2))

    size_x = [n, b.shape[1] + 2 * psf_radius]
    size_z = [n, k, size_x[1]]
    size_k_full = [k, size_x[1]]

    # M is MtM, Mtb is Mtx, the matrix M is zero-padded in 2*psf_radius rows
    # and cols
    M = np.pad(np.ones(b.shape, dtype=real_type),
               ((0, 0), (psf_radius, psf_radius)),
               mode='constant', constant_values=0)
    Mtb = np.pad(b, ((0, 0), (psf_radius, psf_radius)),
                 mode='constant', constant_values=0)

    """Penalty parameters, including the calculation of augmented
       Lagrange multipliers"""
    lambdas = [lambda_residual, lambda_prior]
    gamma_heuristic = 60 * lambda_prior * 1 / np.amax(b)
    gammas_D = [gamma_heuristic / 5000, gamma_heuristic]
    gammas_Z = [gamma_heuristic / 500, gamma_heuristic]
    rho = gammas_D[1] / gammas_D[0]

    """Initialize variables for the d-step"""
    varsize_D = [size_x, size_k_full]
    xi_D = [np.zeros(varsize_D[0], dtype=real_type),
            np.zeros(varsize_D[1], dtype=real_type)]

    xi_D_hat = [np.zeros(varsize_D[0], dtype=imaginary_type),
                np.zeros(varsize_D[1], dtype=imaginary_type)]

    u_D = [np.zeros(varsize_D[0], dtype=real_type),
           np.zeros(varsize_D[1], dtype=real_type)]

    # Lagrange multipliers
    d_D = [np.zeros(varsize_D[0], dtype=real_type),
           np.zeros(varsize_D[1], dtype=real_type)]

    v_D = [np.zeros(varsize_D[0], dtype=real_type),
           np.zeros(varsize_D[1], dtype=real_type)]

    # d = rng.normal(size=size_kernel)
    if ds_init is None:
        d = rng.randn(*size_kernel)
    else:
        d = ds_init.copy()
    d_norm = np.linalg.norm(d, axis=1)
    d /= d_norm[:, None]

    # Initial the filters and its fft after being rolled to fit the frequency
    d = np.pad(d, ((0, 0),
                   (0, size_x[1] - size_kernel[1])),
               mode='constant', constant_values=0)
    d = np.roll(d, -int(psf_radius), axis=1)
    d_hat = fft(d)

    # Initialize variables for the z-step
    varsize_Z = [size_x, size_z]
    xi_Z = [np.zeros(varsize_Z[0], dtype=real_type),
            np.zeros(varsize_Z[1], dtype=real_type)]

    xi_z_hat = [np.zeros(varsize_Z[0], dtype=imaginary_type),
                np.zeros(varsize_Z[1], dtype=imaginary_type)]

    u_Z = [np.zeros(varsize_Z[0], dtype=real_type),
           np.zeros(varsize_Z[1], dtype=real_type)]

    # Lagrange multipliers
    d_Z = [np.zeros(varsize_Z[0], dtype=real_type),
           np.zeros(varsize_Z[1], dtype=real_type)]

    v_Z = [np.zeros(varsize_Z[0], dtype=real_type),
           np.zeros(varsize_Z[1], dtype=real_type)]

    # Initial the codes and its fft
    # z = rng.normal(size=size_z)
    z = np.zeros(size_z)
    z_hat = fft(z)

    """Initial objective function (usually very large)"""
    # obj_val = obj_func(z_hat, d_hat, b,
    #                    lambda_residual, lambda_prior,
    #                    psf_radius, size_z, size_x)
    obj_val = obj_func_2(z, d, b, lambda_prior, psf_radius,
                         feasible_evaluation)

    if verbose > 0:
        print('Init, Obj %3.3f' % (obj_val, ))

    times = list()
    times.append(0.)
    list_obj_val = list()
    list_obj_val.append(obj_val)
    """Start the main algorithm"""
    for i in range(max_it):

        start = time.time()
        z, z_hat = update_z(
            z, z_hat, d_hat, u_Z, v_Z, d_Z, lambdas,
            gammas_Z, Mtb, M, size_x, size_z, xi_Z, xi_z_hat, b,
            lambda_prior, lambda_residual, psf_radius, verbose, max_it_z)
        times.append(time.time() - start)

        # obj_val = obj_func(z_hat, d_hat, b,
        #                    lambda_residual, lambda_prior,
        #                    psf_radius, size_z, size_x)
        obj_val = obj_func_2(z, d, b, lambda_prior, psf_radius,
                             feasible_evaluation)

        if verbose > 0:
            print('Iter Z %d/%d, Obj %3.3f' % (i, max_it, obj_val))

        start = time.time()
        d, d_hat = update_d(
            z_hat, d_hat, size_z, size_x, rho, d,
            v_D, d_D, lambdas, gammas_D, Mtb, u_D, M, size_k_full,
            psf_radius, xi_D, xi_D_hat, verbose, max_it_d)
        times.append(time.time() - start)

        # obj_val = obj_func(z_hat, d_hat, b,
        #                    lambda_residual, lambda_prior,
        #                    psf_radius, size_z, size_x)
        obj_val = obj_func_2(z, d, b, lambda_prior, psf_radius,
                             feasible_evaluation)

        if verbose > 0:
            print('Iter D %d/%d, Obj %3.3f' % (i, max_it, obj_val))

        list_obj_val.append(obj_val)

        # Debug progress
        # z_comp = z

        # Termination
        # if (linalg.norm(z_diff) / linalg.norm(z_comp) < tol and
        #         linalg.norm(d_diff) / linalg.norm(d_comp) < tol):
        #     break
        if stopping_pobj is not None and obj_val < stopping_pobj:
            break

    """Final estimate"""
    z_res = z

    d_res = d
    d_res = np.roll(d_res, psf_radius, axis=1)
    d_res = d_res[:, 0:psf_radius * 2 + 1]

    Dz = np.real(ifft(np.einsum('ijk,jk->ik', z_hat, d_hat)))

    # obj_val = obj_func(z_hat, d_hat, b,
    #                    lambda_residual, lambda_prior,
    #                    psf_radius, size_z, size_x)
    # if verbose > 0:
    #     print('Final objective function %f' % obj_val)
    #
    # reconstr_err = reconstruction_err(z_hat, d_hat, b, psf_radius, size_x)
    # if verbose > 0:
    #     print('Final reconstruction error %f' % reconstr_err)

    return d_res, z_res, Dz, np.array(list_obj_val), times


def KernelConstraintProj(u, size_k_full, psf_radius):
    """Computes the proximal operator for kernel by projection"""

    # Get support
    u_proj = u
    u_proj = np.roll(u_proj, psf_radius, axis=1)
    u_proj = u_proj[:, 0:psf_radius * 2 + 1]

    # Normalize
    u_sum = np.sum(np.power(u_proj, 2), axis=1)
    u_norm = np.tile(u_sum, [u_proj.shape[1], 1]).transpose(1, 0)
    u_proj[u_norm >= 1] = u_proj[u_norm >= 1] / np.sqrt(u_norm[u_norm >= 1])

    # Now shift back and pad again
    u_proj = np.pad(u_proj, ((0, 0),
                             (0, size_k_full[1] - (2 * psf_radius + 1))),
                    mode='constant', constant_values=0)

    u_proj = np.roll(u_proj, -psf_radius, axis=1)

    return u_proj


def precompute_D_step(z_hat, size_z, rho, verbose):
    """Computes to cache the values of Z^.T and (Z^.T*Z^ + rho*I)^-1 as
       in algorithm"""

    n = size_z[0]
    k = size_z[1]

    zhat_mat = np.transpose(z_hat, [2, 0, 1])
    zhat_inv_mat = np.zeros((zhat_mat.shape[0], k, k), dtype=imaginary_type)
    inv_rho_z_hat_z_hat_t = np.zeros(
        (zhat_mat.shape[0], n, n), dtype=imaginary_type)
    z_hat_mat_t = np.transpose(np.ma.conjugate(zhat_mat), [0, 2, 1])

    # Compute z_hat * z_hat^T for each pixel
    z_hat_z_hat_t = np.einsum('knm,kmj->knj', zhat_mat, z_hat_mat_t)

    for i in range(zhat_mat.shape[0]):
        z_hat_z_hat_t_plus_rho = z_hat_z_hat_t[i]
        z_hat_z_hat_t_plus_rho.flat[::n + 1] += rho
        inv_rho_z_hat_z_hat_t[i] = linalg.pinv(z_hat_z_hat_t_plus_rho)

    zhat_inv_mat = 1.0 / rho * (np.eye(k) -
                                np.einsum('knm,kmj->knj',
                                          np.einsum('knm,kmj->knj',
                                                    z_hat_mat_t,
                                                    inv_rho_z_hat_z_hat_t),
                                          zhat_mat))

    # if verbose > 0:
    #     print('Done precomputing for D')
    return zhat_mat, zhat_inv_mat


def precompute_Z_step(dhat, size_x, verbose):
    """Computes to cache the values of D^.T and D^.T*D^ as in algorithm"""

    dhat_flat = dhat.T
    dhatTdhat_flat = np.sum(np.multiply(
        np.ma.conjugate(dhat_flat), dhat_flat), axis=1)
    # if verbose > 0:
    #     print('Done precomputing for Z')
    return dhat_flat, dhatTdhat_flat


def solve_conv_term_D(zhat_mat, zhat_inv_mat, xi_hat, rho, size_z):
    """Solves argmin(||Zd - x1||_2^2 + rho * ||d - x2||_2^2"""

    k = size_z[1]
    sx = size_z[2]

    # Reshape to array per frequency
    xi_hat_0_flat = np.expand_dims(xi_hat[0].T, axis=2)
    xi_hat_1_flat = np.expand_dims(xi_hat[1].T, axis=2)

    x = np.zeros((zhat_mat.shape[0], k), dtype=imaginary_type)
    z_hat_mat_t = np.ma.conjugate(zhat_mat.transpose(0, 2, 1))
    x = np.einsum("ijk, ikl -> ijl", zhat_inv_mat,
                  np.einsum("ijk, ikl -> ijl",
                            z_hat_mat_t,
                            xi_hat_0_flat) +
                  rho * xi_hat_1_flat) \
        .reshape(sx, k)

    # Reshape to get back the new D^
    d_hat = x.T

    return d_hat


def solve_conv_term_Z(dhatT, dhatTdhat, xi_hat, gammas, size_z):
    """Solves argmin(||Dz - x1||_2^2 + rho * ||z - x2||_2^2"""

    sx = size_z[2]
    rho = gammas[1] / gammas[0]

    # Compute b
    xi_hat_0_rep = np.expand_dims(xi_hat[0], axis=1)
    xi_hat_1_rep = xi_hat[1]

    b = (dhatT * xi_hat_0_rep) + rho * xi_hat_1_rep

    scInverse = np.ones((1, sx)) / (rho * np.ones((1, sx)) + dhatTdhat.T)

    dhatT_dot_b = np.ma.conjugate(dhatT) * b
    dhatTb_rep = np.expand_dims(np.sum(dhatT_dot_b, axis=1), axis=1)
    x = 1.0 / rho * (b - (scInverse * dhatT) * dhatTb_rep)

    z_hat = x
    return z_hat


def obj_func(z_hat, d_hat, b, lambda_residual, lambda_prior,
             psf_radius, size_z, size_x):
    """Computes the objective function including the data-fitting and
       regularization terms"""
    # Data-fitting term
    f_z = reconstruction_err(z_hat, d_hat, b, psf_radius, size_x)

    # Regularizer
    z = ifft(z_hat)
    g_z = (lambda_prior / lambda_residual) * np.sum(z)

    f_val = f_z + g_z
    return np.real(f_val)


def reconstruction_err(z_hat, d_hat, b, psf_radius, size_x):
    """Computes the reconstruction error from the data-fitting term"""

    Dz = np.real(ifft(np.einsum('ijk,jk->ik', z_hat, d_hat)))
    err = 0.5 * linalg.norm(Dz[:, psf_radius:-psf_radius] - b) ** 2

    return err


def obj_func_2(z, d, b, lambda_prior, psf_radius, feasible_evaluation=True):
    """Alternative objective function in time domain"""
    d = d.copy()
    z = z.copy()

    z = z[:, :, 2 * psf_radius:-2 * psf_radius]
    z = z.swapaxes(0, 1)

    # get support of d in the time domain
    d = np.roll(d, psf_radius, axis=1)
    d = d[:, 0:psf_radius * 2 + 1]

    if feasible_evaluation:
        # project to unit norm
        d_norm = np.linalg.norm(d, axis=1)
        mask = d_norm >= 1
        d[mask] /= d_norm[mask][:, None]
        # update z in the opposite way
        z[mask] *= d_norm[mask][:, None, None]

    # construct X and compute the objective function with the regularization
    X = construct_X(z, d)
    obj = objective(b, X, z, reg=lambda_prior)
    return obj


def construct_X(Z, ds):
    """
    Parameters
    ----------
    z : array, shape (n_atoms, n_trials, n_times)
        The activations
    ds : array, shape (n_atoms, n_times_atom)
        The atom.

    Returns
    -------
    X : array, shape (n_trials, n_times + n_times_atom - 1)
    """
    assert Z.shape[0] == ds.shape[0]
    n_trials = Z.shape[1]
    n_times_atom = ds.shape[1]
    n_times = Z.shape[2] + n_times_atom - 1
    X = np.zeros((n_trials, n_times))
    for i in range(Z.shape[1]):
        X[i] = sum([signal.convolve(Z[k, i], d)
                    for k, d in enumerate(ds)], 0)
    return X


def objective(X, X_hat, z_hat, reg):
    obj = 0.5 * linalg.norm(X - X_hat, 'fro')**2 + reg * z_hat.sum()
    return obj



"""
==================================================================
Extracting cross-frequency coupling waveforms from rodent LFP data
==================================================================

This example illustrates how to learn univariate atoms on a univariate
time-serie. The data is a single LFP channel recorded on a rodent's striatum
[1]_. Interestingly in this time-serie, the high frequency oscillations around
80 Hz are modulated in amplitude by the low-frequency oscillation around 3 Hz,
a phenomenon known as cross-frequency coupling (CFC).

The convolutional sparse coding (CSC) model is able to learn the prototypical
waveforms of the signal, on which we can clearly see the CFC.

.. [1] G. Dallérac, M. Graupner, J. Knippenberg, R. C. R. Martinez,
    T. F. Tavares, L. Tallot, N. El Massioui, A. Verschueren, S. Höhn,
    J.B. Bertolus, et al. Updating temporal expectancy of an aversive event
    engages striatal plasticity under amygdala control.
    Nature Communications, 8:13920, 2017
"""

# Authors: Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first load the data sample.

import mne
import numpy as np
import matplotlib.pyplot as plt

# sample frequency
sfreq = 350.

# We load the signal. It is an LFP channel recorded on a rodent's striatum.
data = np.load('../rodent_striatum.npy')
print(data.shape)

###############################################################################
# As the data contains severe artifacts between t=0 and t=100, we use a
# section not affected by artifacts.

data = data[:, 35000:]

# We also remove the slow drift, which accounts for a lot of variance.
data = mne.filter.filter_data(data, sfreq, 1, None)

# To make the most of parallel computing, we split the data into trials.
data = data.reshape(50, -1)
data /= data.std()

###############################################################################
# This sample contains CFC between 3 Hz and 80 Hz. This phenomenon can be
# described with a comodulogram, computed for instance with the `pactools
# <http://pactools.github.io/>`_ Python library.

from pactools import Comodulogram

comod = Comodulogram(fs=sfreq, low_fq_range=np.arange(0.2, 10.2, 0.2),
                     low_fq_width=2., method='duprelatour')
comod.fit(data)
comod.plot()
plt.show()

###############################################################################
# We fit a CSC model on the data.

from alphacsc import learn_d_z

params = dict(
    n_atoms=3,
    n_times_atom=int(sfreq * 1.0),  # 1000. ms
    reg=5.,
    n_iter=10,
    solver_z='l-bfgs',
    solver_z_kwargs=dict(factr=1e9),
    solver_d_kwargs=dict(factr=1e2),
    random_state=42,
    n_jobs=5,
    verbose=1)

_, _, d_hat, z_hat, _ = learn_d_z(data, **params)

###############################################################################
# Plot the temporal patterns. Interestingly, we obtain prototypical
# waveforms of the signal on which we can clearly see the CFC.

n_atoms, n_times_atom = d_hat.shape
n_columns = min(6, n_atoms)
n_rows = int(np.ceil(n_atoms // n_columns))
figsize = (4 * n_columns, 3 * n_rows)
fig, axes = plt.subplots(n_rows, n_columns, figsize=figsize, sharey=True)
axes = axes.ravel()

for kk in range(n_atoms):
    ax = axes[kk]
    time = np.arange(n_times_atom) / sfreq
    ax.plot(time, d_hat[kk], color='C%d' % kk)
    ax.set_xlim(0, n_times_atom / sfreq)
    ax.set(xlabel='Time (sec)', title="Temporal pattern %d" % kk)
    ax.grid(True)

fig.tight_layout()
plt.show()



import numpy as np
from scipy import signal


def get_D(uv_hat, n_channels):
    """Compute the rank 1 dictionary associated with the given uv

    Parameter
    ---------
    uv: array (n_atoms, n_channels + n_times_atom)
    n_channels: int
        number of channels in the original multivariate series

    Return
    ------
    D: array (n_atoms, n_channels, n_times_atom)
    """

    return uv_hat[:, :n_channels, None] * uv_hat[:, None, n_channels:]


def get_uv(D):
    """Project D on the space of rank 1 dictionaries

    Parameter
    ---------
    D: array (n_atoms, n_channels, n_times_atom)

    Return
    ------
    uv: array (n_atoms, n_channels + n_times_atom)
    """
    n_atoms, n_channels, n_times_atom = D.shape
    uv = np.zeros((n_atoms, n_channels + n_times_atom))
    for k, d in enumerate(D):
        U, s, V = np.linalg.svd(d)
        uv[k] = np.r_[U[:, 0], V[0]]
    return uv


def _patch_reconstruction_error(X, z, D):
    """Return the reconstruction error for each patches of size (P, L)."""
    n_trials, n_channels, n_times = X.shape
    if D.ndim == 2:
        n_times_atom = D.shape[1] - n_channels
    else:
        n_times_atom = D.shape[2]

    from .convolution import construct_X_multi
    X_hat = construct_X_multi(z, D, n_channels=n_channels)

    diff = (X - X_hat)**2
    patch = np.ones(n_times_atom)

    return np.sum([[np.convolve(patch, diff_ip, mode='valid')
                    for diff_ip in diff_i]
                   for diff_i in diff], axis=1)


def get_lambda_max(X, D_hat, sample_weights=None):

    # univariate case, add a dimension (n_channels = 1)
    if X.ndim == 2:
        X = X[:, None, :]
        D_hat = D_hat[:, None, :]
        if sample_weights is not None:
            sample_weights = sample_weights[:, None, :]

    n_trials, n_channels, n_times = X.shape

    if sample_weights is None:
        # no need for the last dimension if we only have ones
        if D_hat.ndim == 2:
            sample_weights = np.ones(n_trials)
        else:
            sample_weights = np.ones((n_trials, n_channels))

    # multivariate rank-1 case
    if D_hat.ndim == 2:
        return np.max([[
            np.convolve(
                np.dot(uv_k[:n_channels], X_i * W_i), uv_k[:n_channels - 1:-1],
                mode='valid') for X_i, W_i in zip(X, sample_weights)
        ] for uv_k in D_hat], axis=(1, 2))[:, None]

    # multivariate general case
    else:
        return np.max([[
            np.sum([
                np.correlate(D_kp, X_ip * W_ip, mode='valid')
                for D_kp, X_ip, W_ip in zip(D_k, X_i, W_i)
            ], axis=0) for X_i, W_i in zip(X, sample_weights)
        ] for D_k in D_hat], axis=(1, 2))[:, None]


def tukey_window(n_times_atom):
    window = signal.tukey(n_times_atom)
    window[0] = 1e-9
    window[-1] = 1e-9
    return window



# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False

import numpy as np
cimport numpy as cnp
cnp.import_array()

from alphacsc.utils import get_D


cdef _sparse_conv_d(object[:] z_i_data,
                    object[:] z_i_rows,
                    cnp.ndarray[double, ndim=3] D,
                    cnp.ndarray[double, ndim=2] Xi):

    cdef int t
    cdef double ztk
    cdef int n_channels = D.shape[1]
    cdef int n_times_atom = D.shape[2]
    assert n_channels == Xi.shape[0]

    for dk, zk, tk in zip(D, z_i_data, z_i_rows):
        for zkt, t in zip(zk, tk):
            Xi[:, t:t + n_times_atom] += zkt * dk

    return Xi


cdef _sparse_conv_uv(object[:] z_i_data,
                     object[:] z_i_rows,
                     cnp.ndarray[double, ndim=2] u,
                     cnp.ndarray[double, ndim=2] v,
                     cnp.ndarray[double, ndim=2] Xi):

    cdef int t
    cdef double ztk
    cdef cnp.ndarray[double, ndim=1] zik_vk, uk, vk
    cdef int n_channels = u.shape[1]
    cdef int n_times = Xi.shape[1]
    cdef int n_times_atom = v.shape[1]
    assert n_channels == Xi.shape[0]

    for uk, vk, zk, tk in zip(u, v, z_i_data, z_i_rows):
        zik_vk = np.zeros(n_times)
        for zkt, t in zip(zk, tk):
            zik_vk[t:t + n_times_atom] += zkt * vk

        Xi += zik_vk[None, :] * uk[:, None]

    return Xi


def _fast_sparse_convolve_multi(zi_lil, ds):
    n_atoms, n_channels, n_times_atom = ds.shape
    assert zi_lil.shape[0] == n_atoms
    n_atoms, n_times_valid = zi_lil.shape
    n_times = n_times_valid + n_times_atom - 1
    Xi = np.zeros((n_channels, n_times))
    Xi = _sparse_conv_d(zi_lil.data, zi_lil.rows, ds, Xi)
    return Xi


def _fast_sparse_convolve_multi_uv(zi_lil, uv, n_channels, compute_D=True):
    n_atoms, n_channels_n_times_atom = uv.shape
    n_times_atom = n_channels_n_times_atom - n_channels
    assert zi_lil.shape[0] == n_atoms
    n_atoms, n_times_valid = zi_lil.shape
    n_times = n_times_valid + n_times_atom - 1
    Xi = np.zeros((n_channels, n_times))
    if compute_D:
        D = get_D(uv, n_channels)
        return _sparse_conv_d(zi_lil.data, zi_lil.rows, D, Xi)
    else:
        u = uv[:, :n_channels]
        v = uv[:, n_channels:]
        return _sparse_conv_uv(zi_lil.data, zi_lil.rows, u, v, Xi)



import numpy as np

from chainer import Function

from .soft_dtw import SoftDTW
from .distance import SquaredEuclidean


class SoftDTWLoss(Function):

    def __init__(self, gamma):
        self.gamma = gamma

    def forward_cpu(self, inputs):
        # Z, X: both are arrays of shape length x n_dim
        Z, X = inputs

        assert Z.shape[1] == X.shape[1]

        D = SquaredEuclidean(Z, X)
        self.sdtw_ = SoftDTW(D, gamma=self.gamma)
        loss = self.sdtw_.compute()

        return np.array(loss),

    def backward_cpu(self, inputs, grad_outputs):
        Z, X = inputs
        # g has the same shape as the output of forward_cpu().
        # g should always be 1 since it's the last function (loss function)
        g, = grad_outputs

        D = SquaredEuclidean(Z, X)
        E = self.sdtw_.grad()
        gZ = D.jacobian_product(E).astype(Z.dtype)

        # We don't need the gradient w.r.t. the 2nd argument.
        return gZ, np.zeros_like(X)



##########################################################################
# This code is taken from the pactools package available at
# https://github.com/pactools/pactools
#

import numpy as np
from scipy import signal, linalg, fftpack, fft


class Arma(object):
    """Estimator of ARMA model:
        y(t) + a(1)y(t-1) + ... + a(ordar)y(t-ordar) =
        b(0)e(t) + b(1)e(t-1) + ... + b(ordma)e(t-ordma)

    Parameters
    ----------
    ordar : int
        Order of the autoregressive part

    ordma : int
        Order of the moving average part

    block_length : int
        Length of each signal block, on which we estimate the spectrum

    fft_length : int or None
        Length of FFT, should be greater or equal to block_length.
        If None, it is set to block_length

    step : int or None
        Step between successive blocks
        If None, it is set to half the block length (i.e. 0.5 overlap)

    wfunc : function
        Function used to compute the weighting window on each block.
        Examples: np.ones, np.hamming, np.bartlett, np.blackman, ...

    fs : float
        Sampling frequency

    donorm : boolean
        If True, the amplitude is normalized

    """
    def __init__(self, ordar=2, ordma=0, block_length=1024, fft_length=None,
                 step=None, wfunc=np.hamming, fs=1., donorm=True):
        self.ordar = ordar
        self.ordma = ordma
        self.block_length = block_length
        self.fft_length = fft_length
        self.step = step
        self.wfunc = wfunc
        self.fs = fs
        self.donorm = donorm
        self.psd = []

    def check_params(self):
        # block_length
        if self.block_length <= 0:
            raise ValueError('Block length is negative: %s' %
                             (self.block_length, ))
        self.block_length = int(self.block_length)

        # fft_length
        if self.fft_length is None:
            fft_length = next_power2(self.block_length)
        else:
            fft_length = int(self.fft_length)
        if not is_power2(fft_length):
            raise ValueError('FFT length should be a power of 2')
        if fft_length < self.block_length:
            raise ValueError('Block length is greater than FFT length')

        # step
        if self.step is None:
            step = max(int(self.block_length // 2), 1)
        else:
            step = int(self.step)
        if step <= 0 or step > self.block_length:
            raise ValueError('Invalid step between blocks: %s' % (step, ))

        return fft_length, step

    def periodogram(self, signals, hold=False, mean_psd=False):
        """
        Computes the estimation (in dB) for each epoch in a signal

        Parameters
        ----------
        signals : array, shape (n_epochs, n_points)
            Signals from which one computes the power spectrum

        hold : boolean, default = False
            If True, the estimation is appended to the list of previous
            estimations, else, the list is emptied and only the current
            estimation is stored.

        mean_psd : boolean, default = False
            If True, the PSD is the mean PSD over all epochs.

        Returns
        -------
        psd : array, shape (n_epochs, n_freq) or (1, n_freq) if mean_psd
            Power spectrum estimated with a Welsh method on each epoch
            n_freq = fft_length // 2 + 1
        """
        fft_length, step = self.check_params()

        signals = np.atleast_2d(signals)
        n_epochs, n_points = signals.shape

        block_length = min(self.block_length, n_points)

        window = self.wfunc(block_length)
        n_epochs, tmax = signals.shape
        n_freq = fft_length // 2 + 1

        psd = np.zeros((n_epochs, n_freq))

        for i, sig in enumerate(signals):
            block = np.arange(block_length)

            # iterate on blocks
            count = 0
            while block[-1] < sig.size:
                psd[i] += np.abs(
                    fft(window * sig[block], fft_length, 0))[:n_freq] ** 2
                count = count + 1
                block = block + step
            if count == 0:
                raise IndexError(
                    'spectrum: first block has %d samples but sig has %d '
                    'samples' % (block[-1] + 1, sig.size))

            # normalize
            if self.donorm:
                scale = 1.0 / (count * (np.sum(window) ** 2))
            else:
                scale = 1.0 / count
            psd[i] *= scale

        if mean_psd:
            psd = np.mean(psd, axis=0)[None, :]

        if not hold:
            self.psd = []
        self.psd.append(psd)
        return psd

    def estimate(self, nbcorr=np.nan, numpsd=-1):
        fft_length, _ = self.check_params()
        if np.isnan((nbcorr)):
            nbcorr = self.ordar

        # -------- estimate correlation from psd
        full_psd = self.psd[numpsd]
        full_psd = np.c_[full_psd, np.conjugate(full_psd[:, :0:-1])]
        correl = fftpack.ifft(full_psd[0], fft_length, 0).real

        # -------- estimate AR part
        col1 = correl[self.ordma:self.ordma + nbcorr]
        row1 = correl[np.abs(
            np.arange(self.ordma, self.ordma - self.ordar, -1))]
        R = linalg.toeplitz(col1, row1)
        r = -correl[self.ordma + 1:self.ordma + nbcorr + 1]
        AR = linalg.solve(R, r)
        self.AR_ = AR

        # -------- estimate correlation of MA part

        # -------- estimate MA part
        if self.ordma == 0:
            sigma2 = correl[0] + np.dot(AR, correl[1:self.ordar + 1])
            self.MA = np.ones(1) * np.sqrt(sigma2)
        else:
            raise NotImplementedError(
                'arma: estimation of the MA part not yet implemented')

    def arma2psd(self, hold=False):
        """Compute the power spectral density of the ARMA model

        """
        fft_length, _ = self.check_params()
        arpart = np.concatenate((np.ones(1), self.AR_))
        psdar = np.abs(fftpack.fft(arpart, fft_length, 0)) ** 2
        psdma = np.abs(fftpack.fft(self.MA, fft_length, 0)) ** 2
        psd = psdma / psdar
        if not hold:
            self.psd = []
        self.psd.append(psd[None, :fft_length // 2 + 1])

    def inverse(self, sigin):
        """Apply the inverse ARMA filter to a signal

        sigin : input signal (ndarray)

        returns the filtered signal(ndarray)

        """
        arpart = np.concatenate((np.ones(1), self.AR_))
        return signal.fftconvolve(sigin, arpart, 'same')


def ai2ki(ar):
    """Convert AR coefficients to partial correlations
    (inverse Levinson recurrence)

    ar : AR models stored by columns

    returns the partial correlations (one model by column)

    """
    parcor = np.copy(ar)
    ordar, n_epochs, n_points = ar.shape
    for i in range(ordar - 1, -1, -1):
        if i > 0:
            parcor[0:i, :, :] -= (parcor[i:i + 1, :, :] *
                                  np.flipud(parcor[0:i, :, :]))
            parcor[0:i, :, :] *= 1.0 / (1.0 - parcor[i:i + 1, :, :] ** 2)
    return parcor


def ki2ai(parcor):
    """Convert parcor coefficients to autoregressive ones
    (Levinson recurrence)

    parcor : partial correlations stored by columns

    returns the AR models by columns

    """
    ar = np.zeros_like(parcor)
    ordar, n_epochs, n_points = parcor.shape
    for i in range(ordar):
        if i > 0:
            ar[0:i, :, :] += parcor[i:i + 1, :, :] * np.flipud(ar[0:i, :, :])
        ar[i, :, :] = parcor[i, :, :]

    # ok, at least in stationary models
    return ar


def is_power2(num):
    """Test if num is a power of 2. (int -> bool)"""
    num = int(num)
    return num != 0 and ((num & (num - 1)) == 0)


def next_power2(num):
    """Compute the smallest power of 2 >= to num.(float -> int)"""
    return 2 ** int(np.ceil(np.log2(num)))



import warnings
import itertools
from functools import partial

import mne
import numpy as np
import matplotlib.pyplot as plt

from .epoch import make_epochs
from .callback import COLORS, PRINT_KWARGS


def format_arg(arg):
    if isinstance(arg, float):
        return '{:.2f}'.format(arg)
    return arg


def get_label(grid_key, args):
    output = []
    for k in grid_key:
        fmt_string = PRINT_KWARGS.get(k, "{}={{}}".format(k))
        output.append(fmt_string.format(format_arg(args[k])))
    return "\n".join(output)


def _plot_atom(Dk, info, ax, color, plot="atom"):
    sfreq = info.get('sfreq', 1)
    n_channels = info.get('n_channels')
    if plot == "atom":
        t = np.arange(len(Dk) - n_channels) / sfreq
        ax.plot(t, Dk[n_channels:], c=color)
    elif plot == "topo":
        mne.viz.plot_topomap(Dk[:info['n_channels']], info, axes=ax,
                             show=False)
    elif plot == "psd":
        psd = np.abs(np.fft.rfft(Dk[info['n_channels']:])) ** 2
        frequencies = np.linspace(0, sfreq / 2.0, len(psd))
        ax.plot(frequencies, psd, color=color)
        # ax.set(xlabel='Frequencies (Hz)', title='PSD atom final')
        ax.grid(True)
        ax.set_xlim(0, 30)
    else:
        raise NotImplementedError("No plot '{}' for atom".format(plot))


def _plot_activation(z_k, info, ax, color, n_times_atom, t_min=0,
                     plot="density"):
    sfreq = int(info.get('sfreq', 1))
    z_k = z_k.mean(axis=0)  # Average over the epochs
    t = np.arange(len(z_k)) / sfreq + t_min
    if plot == "density":
        blob = np.blackman(n_times_atom)  # bandwidth of n_times_atom
        z_k_smooth = np.convolve(z_k, blob, mode='same')
        ax.fill_between(t, z_k_smooth, color=color, alpha=.5)
    elif plot == "logratio":
        eps = 1e-4
        baseline = z_k[:sfreq]  # Take the first second a a baseline
        mean_baseline = max(baseline.mean(), eps)

        patch = np.ones(sfreq)
        energy = np.maximum(np.convolve(z_k, patch, mode='same'), eps)
        logratio = np.log10(energy / mean_baseline)
        # logratio /= np.std(logratio[:sfreq])
        ax.plot(t[sfreq // 2:-sfreq // 2], logratio[sfreq // 2:-sfreq // 2])
    elif plot == "whiskers":
        # Take 1s at the beginning of the epochs as a baseline
        # Take 1s starting from stim as evoked
        # Take 1s starting 1s after the stim as induced
        baseline = z_k[:sfreq]
        evoked = z_k[(0 < t) * (t < 1)]
        induced = z_k[(1 < t) * (t < 2)]

        ax.boxplot([baseline, evoked, induced], sym='',
                   labels=["baseline", "evoked", "induced"])

        # raise NotImplementedError("Not yet!")
    else:
        raise NotImplementedError("No plot '{}' for activations".format(plot))


def _create_fig(n_cols, n_rows, figname, wrap_col=5, w_plot=5, h_plot=3):
    """Create a figure with subplots

    Parameters
    ----------
    n_cols, n_rows: int, number of rows and column in the plot
    figname: str, name of the figure window
    wrap_col: int, if only 1 row is present, wrap the columns
    w_plot, h_plot: size of the subplots in inches
    """
    wrap_subplots = False
    if n_rows == 1:
        n_rows = n_cols // wrap_col + ((n_cols % wrap_col) != 0)
        n_cols = wrap_col
        wrap_subplots = True

    figsize = (n_cols * w_plot, n_rows * h_plot)

    fig, axes = plt.subplots(n_rows, n_cols, squeeze=False, sharey=True,
                             sharex=True, figsize=figsize, num=figname)
    if wrap_subplots:
        axes = [axes.ravel()]
    return fig, axes


def plot_dictionary(data, info, dirname, name='D_hat', plot="atom"):
    if plot == "topo":
        if not isinstance(info, mne.Info):
            warnings.warn("topomaps should only be used when `info` parameter "
                          "is a subclass of `mne.Info`.", UserWarning)
            return
    n_cols = len(data[0][1][name])
    n_rows = len(data)
    figname = "{}_{}".format(name, plot)
    fig, axes = _create_fig(n_cols, n_rows, figname)

    for row_ax, (arg, res) in zip(axes, data):
        # Reset the color_cycle iterator at each iteration
        color_cycle = itertools.cycle(COLORS)
        for color, ax, Dk in zip(color_cycle, row_ax, res[name]):
            if Dk.ndim == 1:
                # ax.plot(t, Dk[info['n_channels']:], c=color)
                _plot_atom(Dk, info, ax, color, plot=plot)
        row_ax[0].set_ylabel(get_label(info['grid_key'], arg),
                             fontsize=24, labelpad=10, rotation="horizontal",
                             horizontalalignment="right",
                             verticalalignment="center")
        row_ax[0].set_yticks([])
    # fig.subplots_adjust(left=.025, bottom=.025)
    plt.pause(.001)
    fig.tight_layout()
    figname = "{}/{}.png".format(dirname, figname)
    fig.savefig(figname, dpi=150)


def plot_activation(data, info, dirname, name, plot="density"):
    n_cols = len(data[0][1][name])  # n_atoms
    n_rows = len(data)
    figname = "{}_{}".format(name, plot)
    fig, axes = _create_fig(n_cols, n_rows, figname)

    t_min = info.get('t_min', 0)
    for row_ax, (arg, res) in zip(axes, data):
        try:
            z = make_epochs(res[name], info, arg['n_times_atom'])
        except Exception:
            z = res[name]
        color_cycle = itertools.cycle(COLORS)
        # z shape (n_trials, n_atoms, n_times_valid)
        for color, ax, z_k in zip(color_cycle, row_ax, z.swapaxes(0, 1)):
            _plot_activation(z_k, info, ax, color,
                             n_times_atom=arg['n_times_atom'], t_min=t_min,
                             plot=plot)
        row_ax[0].set_ylabel(get_label(info['grid_key'], arg),
                             fontsize=24, labelpad=10, rotation="horizontal",
                             horizontalalignment="right",
                             verticalalignment="center")
        row_ax[0].set_yticks([])
    # fig.subplots_adjust(left=.025, bottom=.025)
    plt.pause(.001)
    fig.tight_layout()
    figname = "{}/{}.png".format(dirname, figname)
    fig.savefig(figname, dpi=150)


def plot_convergence_curve(data, info, dirname):
    # plot the convergence curve
    eps = 1e-6

    # compute the best pobj over all methods
    best_pobj = np.min([np.min(r['pobj']) for _, r in data])

    fig = plt.figure("convergence", figsize=(12, 12))
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

    color_cycle = itertools.cycle(COLORS)
    for (args, res), color in zip(data, color_cycle):
        times = list(np.cumsum(res['times']))
        plt.loglog(
            times, (res['pobj'] - best_pobj) / best_pobj + eps, '.-',
            label=get_label(info['grid_key'], args), color=color,
            linewidth=2)
    plt.xlabel('Time (s)', fontsize=24)
    plt.ylabel('Objective value', fontsize=24)
    ncol = int(np.ceil(len(data) / 10))
    plt.legend(ncol=ncol, fontsize=24)

    plt.gca().tick_params(axis='x', which='both', bottom=False, top=False)
    plt.gca().tick_params(axis='y', which='both', left=False, right=False)
    plt.tight_layout()
    plt.grid(True)
    figname = "{}/convergence.png".format(dirname)
    fig.savefig(figname, dpi=150)


PLOTS = {
    'convergence': plot_convergence_curve,
    'z_init': partial(plot_activation, name='z_init'),
    'z_hat': partial(plot_activation, name='z_hat'),
    'z_hat_whiskers': partial(plot_activation, name='z_hat', plot='whiskers'),
    'z_init_whiskers': partial(plot_activation, name='z_init',
                               plot='whiskers'),
    'z_hat_logratio': partial(plot_activation, name='z_hat', plot='logratio'),
    'z_init_logratio': partial(plot_activation, name='z_init',
                               plot='logratio'),
    'D_init': partial(plot_dictionary, name='D_init'),
    'D_hat': partial(plot_dictionary, name='D_hat'),
    'D_init_psd': partial(plot_dictionary, name='D_init', plot="psd"),
    'D_hat_psd': partial(plot_dictionary, name='D_hat', plot="psd"),
    'D_init_topo': partial(plot_dictionary, name='D_init', plot="topo"),
    'D_hat_topo': partial(plot_dictionary, name='D_hat', plot="topo"),
}


DEFAULT_OUTPUT = [
    'convergence',
    'D_hat_psd', 'D_hat', 'D_init',
    'z_hat', 'z_hat_whiskers'
]



"""Convolutional utilities for dictionary learning"""

# Authors: Thomas Moreau <thomas.moreau@inria.fr>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>

import numba
import numpy as np

from .. import cython_code
from .lil import get_z_shape, is_lil


def construct_X(z, ds):
    """
    Parameters
    ----------
    z : array, shape (n_atoms, n_trials, n_times_valid)
        The activations
    ds : array, shape (n_atoms, n_times_atom)
        The atoms

    Returns
    -------
    X : array, shape (n_trials, n_times)
    """
    assert z.shape[0] == ds.shape[0]
    n_atoms, n_trials, n_times_valid = z.shape
    n_atoms, n_times_atom = ds.shape
    n_times = n_times_valid + n_times_atom - 1

    X = np.zeros((n_trials, n_times))
    for i in range(n_trials):
        X[i] = _choose_convolve(z[:, i], ds)
    return X


def construct_X_multi(z, D=None, n_channels=None):
    """
    Parameters
    ----------
    z : array, shape (n_trials, n_atoms, n_times_valid)
        Can also be a list of n_trials LIL-sparse matrix of shape
            (n_atoms, n_times - n_times_atom + 1)
        The activations
    D : array
        The atoms. Can either be full rank with shape shape
        (n_atoms, n_channels, n_times_atom) or rank 1 with
        shape shape (n_atoms, n_channels + n_times_atom)
    n_channels : int
        Number of channels

    Returns
    -------
    X : array, shape (n_trials, n_channels, n_times)
    """
    n_trials, n_atoms, n_times_valid = get_z_shape(z)
    assert n_atoms == D.shape[0]
    if D.ndim == 2:
        n_times_atom = D.shape[1] - n_channels
    else:
        _, n_channels, n_times_atom = D.shape
    n_times = n_times_valid + n_times_atom - 1

    X = np.zeros((n_trials, n_channels, n_times))
    for i in range(n_trials):
        X[i] = _choose_convolve_multi(z[i], D=D, n_channels=n_channels)
    return X


def _sparse_convolve(z_i, ds):
    """Same as _dense_convolve, but use the sparsity of zi."""
    n_atoms, n_times_atom = ds.shape
    n_atoms, n_times_valid = z_i.shape
    n_times = n_times_valid + n_times_atom - 1
    Xi = np.zeros(n_times)
    for zik, dk in zip(z_i, ds):
        for nnz in np.where(zik != 0)[0]:
            Xi[nnz:nnz + n_times_atom] += zik[nnz] * dk
    return Xi


def _sparse_convolve_multi(z_i, ds):
    """Same as _dense_convolve, but use the sparsity of zi."""
    n_atoms, n_channels, n_times_atom = ds.shape
    n_atoms, n_times_valid = z_i.shape
    n_times = n_times_valid + n_times_atom - 1
    Xi = np.zeros(shape=(n_channels, n_times))
    for zik, dk in zip(z_i, ds):
        for nnz in np.where(zik != 0)[0]:
            Xi[:, nnz:nnz + n_times_atom] += zik[nnz] * dk
    return Xi


def _sparse_convolve_multi_uv(z_i, uv, n_channels):
    """Same as _dense_convolve, but use the sparsity of zi."""
    u = uv[:, :n_channels]
    v = uv[:, n_channels:]
    n_atoms, n_times_valid = z_i.shape
    n_atoms, n_times_atom = v.shape
    n_times = n_times_valid + n_times_atom - 1

    Xi = np.zeros(shape=(n_channels, n_times))
    for zik, uk, vk in zip(z_i, u, v):
        zik_vk = np.zeros(n_times)
        for nnz in np.where(zik != 0)[0]:
            zik_vk[nnz:nnz + n_times_atom] += zik[nnz] * vk

        Xi += zik_vk[None, :] * uk[:, None]

    return Xi


def _dense_convolve(z_i, ds):
    """Convolve z_i[k] and ds[k] for each atom k, and return the sum."""
    return sum([np.convolve(zik, dk) for zik, dk in zip(z_i, ds)], 0)


def _dense_convolve_multi(z_i, ds):
    """Convolve z_i[k] and ds[k] for each atom k, and return the sum."""
    return np.sum([[np.convolve(zik, dkp) for dkp in dk]
                   for zik, dk in zip(z_i, ds)], 0)


def _dense_convolve_multi_uv(z_i, uv, n_channels):
    """Convolve z_i[k] and uv[k] for each atom k, and return the sum."""
    u = uv[:, :n_channels]
    v = uv[:, n_channels:]
    n_atoms, n_times_valid = z_i.shape
    n_atoms, n_times_atom = v.shape
    n_times = n_times_valid + n_times_atom - 1

    Xi = np.zeros((n_channels, n_times))
    for zik, uk, vk in zip(z_i, u, v):
        zik_vk = np.convolve(zik, vk)
        Xi += zik_vk[None, :] * uk[:, None]

    return Xi


def _choose_convolve(z_i, ds):
    """Choose between _dense_convolve and _sparse_convolve with a heuristic
    on the sparsity of z_i, and perform the convolution.

    z_i : array, shape(n_atoms, n_times_valid)
        Activations
    ds : array, shape(n_atoms, n_times_atom)
        Dictionary
    """
    assert z_i.shape[0] == ds.shape[0]

    if np.sum(z_i != 0) < 0.01 * z_i.size:
        return _sparse_convolve(z_i, ds)
    else:
        return _dense_convolve(z_i, ds)


def _choose_convolve_multi(z_i, D=None, n_channels=None):
    """Choose between _dense_convolve and _sparse_convolve with a heuristic
    on the sparsity of z_i, and perform the convolution.

    z_i : array, shape(n_atoms, n_times_valid)
        Activations
    D : array
        The atoms. Can either be full rank with shape shape
        (n_atoms, n_channels, n_times_atom) or rank 1 with
        shape shape (n_atoms, n_channels + n_times_atom)
    n_channels : int
        Number of channels
    """
    assert z_i.shape[0] == D.shape[0]

    if is_lil(z_i):
        cython_code._assert_cython()
        if D.ndim == 2:
            return cython_code._fast_sparse_convolve_multi_uv(
                z_i, D, n_channels, compute_D=True)
        else:
            return cython_code._fast_sparse_convolve_multi(z_i, D)

    elif np.sum(z_i != 0) < 0.01 * z_i.size:
        if D.ndim == 2:
            return _sparse_convolve_multi_uv(z_i, D, n_channels)
        else:
            return _sparse_convolve_multi(z_i, D)

    else:
        if D.ndim == 2:
            return _dense_convolve_multi_uv(z_i, D, n_channels)
        else:
            return _dense_convolve_multi(z_i, D)


@numba.jit((numba.float64[:, :, :], numba.float64[:, :]), cache=True,
           nopython=True)
def numpy_convolve_uv(ztz, uv):
    """Compute the multivariate (valid) convolution of ztz and D

    Parameters
    ----------
    ztz: array, shape = (n_atoms, n_atoms, 2 * n_times_atom - 1)
        Activations
    uv: array, shape = (n_atoms, n_channels + n_times_atom)
        Dictionnary

    Returns
    -------
    G : array, shape = (n_atoms, n_channels, n_times_atom)
        Gradient
    """
    assert uv.ndim == 2
    n_times_atom = (ztz.shape[2] + 1) // 2
    n_atoms = ztz.shape[0]
    n_channels = uv.shape[1] - n_times_atom

    u = uv[:, :n_channels]
    v = uv[:, n_channels:][:, ::-1]

    G = np.zeros((n_atoms, n_channels, n_times_atom))
    for k0 in range(n_atoms):
        for k1 in range(n_atoms):
            for t in range(n_times_atom):
                G[k0, :, t] += (
                    np.sum(ztz[k0, k1, t:t + n_times_atom] * v[k1]) * u[k1, :])

    return G


def tensordot_convolve(ztz, D):
    """Compute the multivariate (valid) convolution of ztz and D

    Parameters
    ----------
    ztz: array, shape = (n_atoms, n_atoms, 2 * n_times_atom - 1)
        Activations
    D: array, shape = (n_atoms, n_channels, n_times_atom)
        Dictionnary

    Returns
    -------
    G : array, shape = (n_atoms, n_channels, n_times_atom)
        Gradient
    """
    n_atoms, n_channels, n_times_atom = D.shape
    D_revert = D[:, :, ::-1]

    G = np.zeros(D.shape)
    for t in range(n_times_atom):
        G[:, :, t] = np.tensordot(ztz[:, :, t:t + n_times_atom], D_revert,
                                  axes=([1, 2], [0, 2]))
    return G


def sort_atoms_by_explained_variances(D_hat, z_hat, n_channels):
    n_atoms = D_hat.shape[0]
    assert z_hat.shape[1] == n_atoms
    variances = np.zeros(n_atoms)
    for kk in range(n_atoms):
        variances[kk] = construct_X_multi(z_hat[:, kk:kk + 1],
                                          D_hat[kk:kk + 1],
                                          n_channels=n_channels).var()
    order = np.argsort(variances)[::-1]
    z_hat = z_hat[:, order, :]
    D_hat = D_hat[order, ...]
    return D_hat, z_hat



"""
=============================
Vanilla CSC on simulated data
=============================

This example demonstrates `vanilla` CSC on simulated data.
Note that `vanilla` CSC is just a special case of alphaCSC with alpha=2.
"""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first define the parameters of our model.

n_times_atom = 64  # L
n_times = 512  # T
n_atoms = 2  # K
n_trials = 100  # N
n_iter = 50

reg = 0.1

###############################################################################
# Here, we simulate the data

from alphacsc.simulate import simulate_data # noqa

random_state_simulate = 1
X, ds_true, z_true = simulate_data(n_trials, n_times, n_times_atom,
                                   n_atoms, random_state_simulate)

###############################################################################
# Add some noise and corrupt some trials

from scipy.stats import levy_stable # noqa
from alphacsc import check_random_state # noqa

# Add stationary noise:
fraction_corrupted = 0.02
n_corrupted_trials = int(fraction_corrupted * n_trials)

rng = check_random_state(random_state_simulate)
X += 0.01 * rng.randn(*X.shape)

idx_corrupted = rng.randint(0, n_trials,
                            size=n_corrupted_trials)

###############################################################################
# Let us look at the first 10 trials to see how they look.

from alphacsc.viz.callback import plot_data # noqa
plot_data([X[:10]])

###############################################################################
# Note that the atoms don't always have the same amplitude or occur at the
# same time instant.
#
# Now, we run vanilla CSC on the data.

from alphacsc import learn_d_z # noqa

random_state = 60

pobj, times, d_hat, z_hat, reg = learn_d_z(
    X, n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
    solver_d_kwargs=dict(factr=100), random_state=random_state,
    n_jobs=1, verbose=1)
print('Vanilla CSC')

###############################################################################
# Finally, let's compare the results.

import matplotlib.pyplot as plt # noqa
plt.figure()
plt.plot(d_hat.T)
plt.plot(ds_true.T, 'k--')

###############################################################################
# We can also visualize the learned activations

plot_data([z[:10] for z in z_hat], ['stem'] * n_atoms)

###############################################################################
# Note if the data is corrupted with impulsive noise, this method may not be
# the best.  Check out our :ref:`example using alphacsc
# <sphx_glr_auto_examples_plot_simulate_alphacsc.py>` to learn how to deal with
# such data.

alpha = 1.2
noise_level = 0.005
X[idx_corrupted] += levy_stable.rvs(alpha, 0, loc=0, scale=noise_level,
                                    size=(n_corrupted_trials, n_times),
                                    random_state=random_state_simulate)
pobj, times, d_hat, z_hat, reg = learn_d_z(
    X, n_atoms, n_times_atom, reg=reg, n_iter=n_iter,
    solver_d_kwargs=dict(factr=100), random_state=random_state,
    n_jobs=1, verbose=1)
plt.figure()
plt.plot(d_hat.T)
plt.plot(ds_true.T, 'k--')
plt.show()



import re
import os
import hcp
import mne
import numpy as np
from glob import glob
from copy import deepcopy
from joblib import Memory
from hcp.io.file_mapping.file_mapping import kind_map

from ..utils import check_random_state


HCP_DIR = "/storage/store/data/HCP900/"
CONVERSION_MAP = {v: k for k, v in kind_map.items()}


mem = Memory(location='.', verbose=0)


def get_all_records(hcp_path=HCP_DIR):
    """Make a dictionary with all HCP files in the directory hcp_path

    Parameters
    ----------
    hcp_path: str
        Path in which the HCP files are located

    Return
    ------
    db: dict
        Dictionary with {data_type: {subject: [run_index_0, run_index_1, ...]}}
        The keys are str for the type of exercises and the values are
        dictionaries containing a list per subject with the run_indexes.

    """
    # List all files with unprocesses
    pattern = os.path.join(hcp_path, "*/unprocessed/MEG/*/4D/config")
    list_files = glob(pattern)
    db = {}
    pattern = pattern.replace("*", "(.*)")
    pattern = pattern.replace("MEG/", "MEG/\d+-")  # noqa
    for f_name in list_files:
        subject, data_type = re.match(pattern, f_name).groups()
        data_type = CONVERSION_MAP[data_type]
        type_subjects = db.get(data_type, {})
        type_subject_records = type_subjects.get(subject, [])
        type_subject_records += [len(type_subject_records)]
        type_subjects[subject] = type_subject_records
        db[data_type] = type_subjects
    print("Found {} types".format(len(db.keys())))

    return db


@mem.cache(ignore=['n_jobs'])
def load_one_record(data_type, subject, run_index, sfreq=300, epoch=None,
                    filter_params=[5., None], n_jobs=1):
    # Load the record and correct the sensor space to get proper visualization
    print(f"subject={subject}, data_type={data_type}, run_index={run_index}, "
          f"hcp_path={HCP_DIR}")
    raw = hcp.read_raw(subject, data_type=data_type, run_index=run_index,
                       hcp_path=HCP_DIR, verbose=0)
    raw.load_data()
    hcp.preprocessing.map_ch_coords_to_mne(raw)
    raw.pick_types(meg='mag', eog=False, stim=True)

    # filter the electrical and low frequency components
    raw.notch_filter([60, 120], n_jobs=n_jobs)
    raw.filter(*filter_params, n_jobs=n_jobs)

    # Resample to the requested sfreq
    if sfreq is not None:
        raw.resample(sfreq=sfreq, n_jobs=n_jobs)

    events = mne.find_events(raw)
    raw.pick_types(meg='mag', stim=False)
    events[:, 0] -= raw.first_samp

    # Deep copy before modifying info to avoid issues when saving EvokedArray
    info = deepcopy(raw.info)
    info['events'] = events
    info['event_id'] = np.unique(events[:, 2])

    # Return the data
    return raw.get_data(), info


def load_data(n_trials=10, data_type='rest', sfreq=150, epoch=None,
              filter_params=[5., None], equalize="zeropad", n_jobs=1,
              random_state=None):
    """Load and prepare the HCP dataset for multiCSC


    Parameters
    ----------
    n_trials : int
        Number of recordings that are loaded.
    data_type : str
        Type of recordings loaded. Should be in {'rest', 'task_working_memory',
        'task_motor', 'task_story_math', 'noise_empty_room', 'noise_subject'}.
    sfreq : float
        Sampling frequency of the signal. The data are resampled to match it.
    epoch : tuple or None
        If set to a tuple, extract epochs from the raw data, using
        t_min=epoch[0] and t_max=epoch[1]. Else, use the raw signal, divided
        in n_splits chunks.
    filter_params : tuple
        Frequency cut for a band pass filter applied to the signals. The
        default is a high-pass filter with frequency cut at 2Hz.
    n_jobs : int
        Number of jobs that can be used for preparing (filtering) the data.
    random_state : int | None
        State to seed the random number generator.

    Return
    ------
    X : ndarray, shape (n_trials, n_channels, n_times)
        Signals loaded from HCP.
    info : list of mne.Info
        List of the info related to each signals.
    """
    if data_type == "rest" and epoch is not None:
        raise ValueError("epoch != None is not valid with resting-state data.")

    rng = check_random_state(random_state)
    mne.set_log_level(30)

    db = get_all_records()
    records = [(subject, run_index)
               for subject, runs in db[data_type].items()
               for run_index in runs]

    X, info = [], []
    records = rng.permutation(records)[:n_trials]
    for i, (subject, run_index) in enumerate(records):
        print("\rLoading HCP subjects: {:7.2%}".format(i / n_trials),
              end='', flush=True)
        X_n, info_n = load_one_record(
            data_type, subject, int(run_index), sfreq=sfreq, epoch=epoch,
            filter_params=filter_params, n_jobs=n_jobs)
        X += [X_n]
        info += [info_n]

    print("\rLoading HCP subjects: done   ")
    X = make_array(X, equalize=equalize)
    X /= np.std(X)
    return X, info


def data_generator(n_trials=10, data_type='rest', sfreq=150, epoch=None,
                   filter_params=[5., None], equalize="zeropad", n_jobs=1,
                   random_state=None):
    """Generator loading subjects from the HCP dataset for multiCSC


    Parameters
    ----------
    n_trials : int
        Number of recordings that are loaded.
    data_type : str
        Type of recordings loaded. Should be in {'rest', 'task_working_memory',
        'task_motor', 'task_story_math', 'noise_empty_room', 'noise_subject'}.
    sfreq : float
        Sampling frequency of the signal. The data are resampled to match it.
    epoch : tuple or None
        If set to a tuple, extract epochs from the raw data, using
        t_min=epoch[0] and t_max=epoch[1]. Else, use the raw signal, divided
        in n_splits chunks.
    filter_params : tuple
        Frequency cut for a band pass filter applied to the signals. The
        default is a high-pass filter with frequency cut at 2Hz.
    n_jobs : int
        Number of jobs that can be used for preparing (filtering) the data.
    random_state : int | None
        State to seed the random number generator.

    Yields
    ------
    X : ndarray, shape (1, n_channels, n_times)
        Signals loaded from HCP.
    info : list of mne.Info
        info related to this signal.
    """
    if data_type == "rest" and epoch is not None:
        raise ValueError("epoch != None is not valid with resting-state data.")

    rng = check_random_state(random_state)
    mne.set_log_level(30)

    db = get_all_records()
    records = [(subject, run_index)
               for subject, runs in db[data_type].items()
               for run_index in runs]

    records = rng.permutation(records)[:n_trials]
    for i, (subject, run_index) in enumerate(records):
        try:
            X_n, info_n = load_one_record(
                data_type, subject, int(run_index), sfreq=sfreq, epoch=epoch,
                filter_params=filter_params, n_jobs=n_jobs)
            X_n /= X_n.std()
            yield X_n, info_n
        except UnicodeDecodeError:
            print("failed to load {}-{}-{}"
                  .format(subject, data_type, run_index))


def make_array(X, equalize='zeropad'):
    """"""
    x_shape = np.array([x.shape for x in X])
    assert np.all(x_shape[..., :-1] == x_shape[0, ..., :-1])
    if equalize == "crop":
        L = x_shape.min(axis=0)[-1]
        X = np.array([x[..., :L] for x in X])
    elif equalize == "zeropad":
        X_shape = tuple(x_shape.max(axis=0))
        X_shape, L = X_shape[:-1], X_shape[-1]
        X = np.array([
            np.concatenate([x, np.zeros(X_shape + (L - x.shape[-1], ))],
                           axis=-1) for x in X
        ])
    else:
        raise ValueError("The equalize '{}' is not valid. It should be in "
                         "{'crop', 'zeropad'}".format(equalize))

    return X



"""
===========================================================
Extracting :math:`\mu`-wave from the somato-sensory dataset
===========================================================

This example illustrates how to learn rank-1 atoms [1]_ on the multivariate
somato-sensorymotor dataset from :code:`mne`. The displayed results highlight
the presence of :math:`\mu`-waves located in the SI cortex.

.. [1] Dupré La Tour, T., Moreau, T., Jas, M., & Gramfort, A. (2018).
    `Multivariate Convolutional Sparse Coding for Electromagnetic Brain Signals
    <https://arxiv.org/abs/1805.09654v2>`_. Advances in Neural Information
    Processing Systems (NIPS).
"""

# Authors: Thomas Moreau <thomas.moreau@inria.fr>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first define the parameters of our model.

sfreq = 150.

# Define the shape of the dictionary
n_atoms = 25
n_times_atom = int(round(sfreq * 1.0))  # 1000. ms

###############################################################################
# Next, we define the parameters for multivariate CSC

from alphacsc import BatchCDL
cdl = BatchCDL(
    # Shape of the dictionary
    n_atoms=n_atoms,
    n_times_atom=n_times_atom,
    # Request a rank1 dictionary with unit norm temporal and spatial maps
    rank1=True, uv_constraint='separate',
    # Initialize the dictionary with random chunk from the data
    D_init='chunk',
    # rescale the regularization parameter to be 20% of lambda_max
    lmbd_max="scaled", reg=.2,
    # Number of iteration for the alternate minimization and cvg threshold
    n_iter=100, eps=1e-4,
    # solver for the z-step
    solver_z="lgcd", solver_z_kwargs={'tol': 1e-2, 'max_iter': 1000},
    # solver for the d-step
    solver_d='alternate_adaptive', solver_d_kwargs={'max_iter': 300},
    # Technical parameters
    verbose=1, random_state=0, n_jobs=6)


###############################################################################
# Here, we load the data from the somato-sensory dataset and preprocess them
# in epochs. The epochs are selected around the stim, starting 2 seconds
# before and finishing 4 seconds after.

from alphacsc.datasets.somato import load_data
t_lim = (-2, 4)
X, info = load_data(epoch=t_lim, sfreq=sfreq)


###############################################################################
# Fit the model and learn rank1 atoms
cdl.fit(X)

###############################################################################
# Display the 4-th atom, which displays a :math:`\mu`-waveform in its temporal
# pattern.

import mne
import numpy as np
import matplotlib.pyplot as plt

i_atom = 4
n_plots = 3
figsize = (n_plots * 3.5, 5)
fig, axes = plt.subplots(1, n_plots, figsize=figsize, squeeze=False)

# Plot the spatial map of the learn atom using mne topomap
ax = axes[0, 0]
u_hat = cdl.u_hat_[i_atom]
mne.viz.plot_topomap(u_hat, info, axes=ax, show=False)
ax.set(title='Learned spatial pattern')

# Plot the temporal pattern of the learn atom
ax = axes[0, 1]
v_hat = cdl.v_hat_[i_atom]
t = np.arange(v_hat.size) / sfreq
ax.plot(t, v_hat)
ax.set(xlabel='Time (sec)', title='Learned temporal waveform')
ax.grid(True)

# Plot the psd of the time atom
ax = axes[0, 2]
psd = np.abs(np.fft.rfft(v_hat)) ** 2
frequencies = np.linspace(0, sfreq / 2.0, len(psd))
ax.semilogy(frequencies, psd)
ax.set(xlabel='Frequencies (Hz)', title='Power Spectral Density')
ax.grid(True)
ax.set_xlim(0, 30)

plt.tight_layout()
plt.show()



import time

import numpy as np
from scipy import optimize

from .compute_constants import compute_DtD
from . import check_random_state


def _support_least_square(X, uv, z, debug=False):
    """WIP, not fonctional!"""
    n_trials, n_channels, n_times = X.shape
    n_atoms, _, n_times_valid = z.shape
    n_times_atom = n_times - n_times_valid + 1

    # Compute DtD
    DtD = compute_DtD(uv, n_channels)
    t0 = n_times_atom - 1
    z_hat = np.zeros(z.shape)

    for idx in range(n_trials):
        Xi = X[idx]
        support_i = z[:, idx].nonzero()
        n_support = len(support_i[0])
        if n_support == 0:
            continue
        rhs = np.zeros((n_support, n_support))
        lhs = np.zeros(n_support)

        for i, (k_i, t_i) in enumerate(zip(*support_i)):
            for j, (k_j, t_j) in enumerate(zip(*support_i)):
                dt = t_i - t_j
                if abs(dt) < n_times_atom:
                    rhs[i, j] = DtD[k_i, k_j, t0 + dt]
            aux_i = np.dot(uv[k_i, :n_channels], Xi[:, t_i:t_i + n_times_atom])
            lhs[i] = np.dot(uv[k_i, n_channels:], aux_i)

        # Solve the non-negative least-square with nnls
        z_star, a = optimize.nnls(rhs, lhs)
        for i, (k_i, t_i) in enumerate(zip(*support_i)):
            z_hat[k_i, idx, t_i] = z_star[i]

    return z_hat


def fista(f_obj, f_grad, f_prox, step_size, x0, max_iter, verbose=0,
          momentum=False, eps=None, adaptive_step_size=False, debug=False,
          scipy_line_search=True, name='ISTA', timing=False, restart=None):
    """ISTA and FISTA algorithm

    Parameters
    ----------
    f_obj : callable
        Objective function. Used only if debug or adaptive_step_size.
    f_grad : callable
        Gradient of the objective function
    f_prox : callable
        Proximal operator
    step_size : float or None
        Step size of each update. Can be None if adaptive_step_size.
    x0 : array
        Initial point of the optimization
    max_iter : int
        Maximum number of iterations
    verbose : int
        Verbosity level
    momentum : boolean
        If True, use FISTA instead of ISTA
    eps : float or None
        Tolerance for the stopping criterion
    adaptive_step_size : boolean
        If True, the step size is adapted at each step
    debug : boolean
        If True, compute the objective function at each step and return the
        list at the end.
    timing : boolean
        If True, compute the objective function at each step, and the duration
        of each step, and return both lists at the end.
    restart : int or None
        If not None, restart the momentum every `restart` iterations.

    Returns
    -------
    x_hat : array
        The final point after optimization
    pobj : list or None
        If debug is True, pobj contains the value of the cost function at each
        iteration.
    """
    pobj = None
    if debug:
        pobj = list()
    if timing:
        times = [0]
        pobj = [f_obj(x0)]
        start = time.time()

    if step_size is None:
        step_size = 1.
    if eps is None:
        eps = np.finfo(np.float32).eps
    obj_uv = None

    tk = 1.0
    x_hat = x0.copy()
    x_hat_aux = x_hat.copy()
    grad = np.empty(x_hat.shape)
    diff = np.empty(x_hat.shape)
    for ii in range(max_iter):
        # restart every n iterations
        if restart is not None and ii > 0 and (ii % restart) == 0:
            x_hat_aux = x_hat.copy()
            tk = 1.0

        grad[:] = f_grad(x_hat_aux)

        if adaptive_step_size:
            if scipy_line_search:

                def f_obj_(x_hat):
                    x_hat = np.reshape(x_hat, x0.shape)
                    return f_obj(f_prox(x_hat))

                step_size, _, obj_uv = optimize.linesearch.line_search_armijo(
                    f_obj_, x_hat.ravel(), -grad.ravel(), grad.ravel(), obj_uv,
                    c1=1e-5, alpha0=step_size)
                if step_size is None:
                    step_size = 0
                x_hat_aux -= step_size * grad
                x_hat_aux = f_prox(x_hat_aux)

            else:

                def f(step_size):
                    x_hat = f_prox(x_hat_aux - step_size * grad)
                    pobj = f_obj(x_hat)
                    return pobj, x_hat

                obj_uv, x_hat_aux, step_size = _adaptive_step_size(
                    f, obj_uv, alpha=step_size)

        else:
            x_hat_aux -= step_size * grad
            x_hat_aux = f_prox(x_hat_aux)

        diff[:] = x_hat_aux - x_hat
        x_hat[:] = x_hat_aux
        if momentum:
            tk_new = (1 + np.sqrt(1 + 4 * tk * tk)) / 2
            x_hat_aux += (tk - 1) / tk_new * diff
            tk = tk_new

        if debug:
            pobj.append(f_obj(x_hat))
        if timing:
            times.append(time.time() - start)
            pobj.append(f_obj(x_hat))
            start = time.time()

        f = np.sum(abs(diff))
        if f <= eps:
            break
        if f > 1e50:
            raise RuntimeError("The D update have diverged.")
    else:
        if verbose > 1:
            print('[{}] update did not converge'.format(name))
    if verbose > 5:
        print('[{}]: {} iterations'.format(name, ii + 1))

    if timing:
        return x_hat, pobj, times
    return x_hat, pobj


def _adaptive_step_size(f, f0=None, alpha=None, tau=2):
    """
    Parameters
    ----------
    f : callable
        Optimized function, take only the step size as argument
    f0 : float
        value of f at current point, i.e. step size = 0
    alpha : float
        Initial step size
    tau : float
        Multiplication factor of the step size during the adaptation
    """

    if alpha is None:
        alpha = 1

    if f0 is None:
        f0, _ = f(0)
    f_alpha, x_alpha = f(alpha)
    f_alpha_down, x_alpha_down = f(alpha / tau)
    f_alpha_up, x_alpha_up = f(alpha * tau)

    alphas = [0, alpha / tau, alpha, alpha * tau]
    fs = [f0, f_alpha_down, f_alpha, f_alpha_up]
    xs = [None, x_alpha_down, x_alpha, x_alpha_up]
    i = np.argmin(fs)
    if i == 0:
        alpha /= tau * tau
        f_alpha, x_alpha = f(alpha)
        while f0 <= f_alpha and alpha > 1e-20:
            alpha /= tau
            f_alpha, x_alpha = f(alpha)
        return f_alpha, x_alpha, alpha
    else:
        return fs[i], xs[i], alphas[i]


def power_iteration(lin_op, n_points=None, b_hat_0=None, max_iter=1000,
                    tol=1e-7, random_state=None):
    """Estimate dominant eigenvalue of linear operator A.

    Parameters
    ----------
    lin_op : callable or array
        Linear operator from which we estimate the largest eigenvalue.
    n_points : tuple
        Input shape of the linear operator `lin_op`.
    b_hat_0 : array, shape (n_points, )
        Init vector. The estimated eigen-vector is stored inplace in `b_hat_0`
        to allow warm start of future call of this function with the same
        variable.

    Returns
    -------
    mu_hat : float
        The largest eigenvalue
    """
    if hasattr(lin_op, 'dot'):
        n_points = lin_op.shape[1]
        lin_op = lin_op.dot
    elif callable(lin_op):
        msg = ("power_iteration require n_points argument when lin_op is "
               "callable")
        assert n_points is not None, msg
    else:
        raise ValueError("lin_op should be a callable or a ndarray")

    rng = check_random_state(random_state)
    if b_hat_0 is None:
        b_hat = rng.rand(n_points)
    else:
        b_hat = b_hat_0

    mu_hat = np.nan
    for ii in range(max_iter):
        b_hat = lin_op(b_hat)
        norm = np.linalg.norm(b_hat)
        if norm == 0:
            return 0
        b_hat /= norm
        fb_hat = lin_op(b_hat)
        mu_old = mu_hat
        mu_hat = np.dot(b_hat, fb_hat)
        # note, we might exit the loop before b_hat converges
        # since we care only about mu_hat converging
        if (mu_hat - mu_old) / mu_old < tol:
            break

    assert not np.isnan(mu_hat)

    if b_hat_0 is not None:
        # copy inplace into b_hat_0 for next call to power_iteration
        np.copyto(b_hat_0, b_hat)

    return mu_hat



import numpy as np
import matplotlib.pyplot as plt
from scipy import signal

from .arma import Arma


def whitening(X, ordar=10, block_length=256, sfreq=1., zero_phase=True,
              plot=False, use_fooof=False):
    n_trials, n_channels, n_times = X.shape

    ar_model = Arma(ordar=ordar, ordma=0, fs=sfreq, block_length=block_length)
    ar_model.periodogram(X.reshape(-1, n_times), hold=False, mean_psd=True)

    if use_fooof:  # pragma: no cover
        # Fit the psd with a 1/f^a background model plus a gaussian mixture.
        # We keep only the background model
        # (pip install fooof)
        from fooof import FOOOF
        fm = FOOOF(background_mode='fixed', verbose=False)
        power_spectrum = ar_model.psd[-1][0]
        freqs = np.linspace(0, sfreq / 2.0, len(power_spectrum))

        fm.fit(freqs, power_spectrum, freq_range=None)
        # repete first point, which is f_0
        bg_fit = np.r_[fm._bg_fit[0], fm._bg_fit][None, :]
        ar_model.psd.append(np.power(10, bg_fit))

    if zero_phase:
        ar_model.psd[-1] = np.sqrt(ar_model.psd[-1])
    ar_model.estimate()

    # apply the whitening for zero-phase filtering
    X_white = apply_whitening(ar_model, X, zero_phase=zero_phase, mode='same')
    assert X_white.shape == X.shape

    # removes edges
    n_times_white = X_white.shape[-1]
    X_white *= signal.tukey(n_times_white,
                            alpha=3 / float(n_times_white))[None, None, :]

    if plot:  # pragma: no cover
        # plot the Power Spectral Density (PSD) before/after
        ar_model.arma2psd(hold=True)
        if zero_phase:
            ar_model.psd[-2] **= 2
            ar_model.psd[-1] **= 2
        ar_model.periodogram(
            X_white.reshape(-1, n_times), hold=True, mean_psd=True)
        labels = ['signal', 'model AR', 'signal white']
        if use_fooof:
            labels = ['signal', 'FOOOF fit', 'model AR', 'signal white']
        ar_model.plot('periodogram before/after whitening',
                      labels=labels, fscale='lin')
        plt.legend(loc='lower left')

    return ar_model, X_white


def apply_ar(ar_model, x, zero_phase=True, mode='same', reverse_ar=False):
    # TODO: speed-up with only one conv
    ar_coef = np.concatenate((np.ones(1), ar_model.AR_))
    if reverse_ar:
        ar_coef = ar_coef[::-1]

    if zero_phase:
        tmp = signal.fftconvolve(x, ar_coef, mode)[::-1]
        return signal.fftconvolve(tmp, ar_coef, mode)[::-1]
    else:
        return signal.fftconvolve(x, ar_coef, mode)


def apply_whitening(ar_model, X, zero_phase=True, mode='same',
                    reverse_ar=False, n_channels=None):
    if X.ndim == 2:
        msg = "For rank1 D, n_channels should be provided"
        assert n_channels is not None, msg
        v = X[:, n_channels:]
        v_white = [apply_ar(ar_model, vk, zero_phase=zero_phase,
                            mode=mode) for vk in v]
        return np.c_[X[:, :n_channels], v_white]

    elif X.ndim == 3:
        return np.array([[
            apply_ar(ar_model, Xij, zero_phase=zero_phase, mode=mode,
                     reverse_ar=reverse_ar) for Xij in Xi]
            for Xi in X])
    else:
        raise NotImplementedError("Should not be called!")


def unwhitening(ar_model, X_white, estimate=True, zero_phase=True, plot=False):
    if X_white.ndim == 2:
        X_white = X_white[None, :]

    n_trials, n_channels, n_times = X_white.shape

    if estimate:
        ar_model.arma2psd(hold=True)
        ar_model.psd[-1] = 1. / ar_model.psd[-1]

        ar_model.estimate()

    # apply the whitening twice (forward and backward) for zero-phase filtering
    X_unwhite = apply_whitening(ar_model, X_white, zero_phase=zero_phase)
    assert X_unwhite.shape == X_white.shape

    if plot:
        # plot the Power Spectral Density (PSD) before/after
        ar_model.periodogram(
            X_white.reshape(-1, n_times), hold=False, mean_psd=True)
        ar_model.arma2psd(hold=True)
        if zero_phase:
            ar_model.psd[-1] **= 2
        ar_model.periodogram(
            X_unwhite.reshape(-1, n_times), hold=True, mean_psd=True)
        ar_model.plot('periodogram before/after unwhitening',
                      labels=['signal white', 'model AR',
                              'signal unwhite'], fscale='lin')
        plt.legend(loc='lower left')
    return X_unwhite



"""
=====================
SWM on simulated data
=====================

This example shows how the sliding window method (SWM) [1]
works on simulated data. The code is adapted from the
`neurodsp package <https://github.com/voytekresearch/neurodsp/>`_
from Voytek lab. Note that, at present, it does not
implement parallel tempering.

[1] Gips, Bart, et al.
    Discovering recurring patterns in electrophysiological recordings.
    Journal of neuroscience methods 275 (2017): 66-79.
"""

# Authors: Scott Cole
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us define the model parameters

n_times_atom = 64  # L
n_times = 5000  # T
n_trials = 10  # N

###############################################################################
# The algorithm does not naturally lend itself to multiple atoms. Therefore,
# we simulate only one atom.
n_atoms = 1  # K

###############################################################################
# A minimum spacing between the windows averaged must be found.
min_spacing = 200  # G

###############################################################################
# Now, we can simulate
from alphacsc import check_random_state
from alphacsc.simulate import simulate_data

random_state_simulate = 1
X, ds_true, z_true = simulate_data(n_trials, n_times, n_times_atom,
                                   n_atoms, random_state_simulate,
                                   constant_amplitude=True)

rng = check_random_state(random_state_simulate)
X += 0.01 * rng.randn(*X.shape)

###############################################################################
# We expect 10 occurences of the atom in total.
# So, let us define 10 random locations for the algorithm to start with.
# If this number is not known, we will end up estimating more/less windows.
import numpy as np
window_starts = rng.choice(np.arange(n_trials * n_times), size=n_trials)

###############################################################################
# Now, we apply the SWM algorithm now.
from alphacsc.other.swm import sliding_window_matching

random_state = 42
X = X.reshape(X.shape[0] * X.shape[1])  # expects 1D time series
d_hat, window_starts, J = sliding_window_matching(
    X, L=n_times_atom, G=min_spacing, window_starts_custom=window_starts,
    max_iterations=10000, T=0.01, random_state=random_state)

###############################################################################
# Let us look at the data at the time windows when the atoms are found.
import matplotlib.pyplot as plt
fig, axes = plt.subplots(2, n_trials // 2, sharex=True, sharey=True,
                         figsize=(15, 3))
axes = axes.ravel()
for ax, w_start in zip(axes, window_starts):
    ax.plot(X[w_start:w_start + n_times_atom])

###############################################################################
# It is not perfect, but it does find time windows where the atom
# is present. Now let us plot the atoms.

plt.figure()
plt.plot(d_hat / np.linalg.norm(d_hat))
plt.plot(ds_true.T, '--')

###############################################################################
# and the cost function over iterations

plt.figure()
plt.plot(J)
plt.ylabel('Cost function J')
plt.xlabel('Iteration #')
plt.show()
